from .convergence import plot_convergence, plot_block_average
from .dF_state import plot_dF_state
from .mbar_matrix import plot_mbar_overlap_matrix
from .ti_dhdl import plot_ti_dhdl

__all__ = [
    "plot_convergence",
    "plot_block_average",
    "plot_dF_state",
    "plot_mbar_overlap_matrix",
    "plot_ti_dhdl",
]



import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator

from .base import _EstimatorMixOut


class TI(BaseEstimator, _EstimatorMixOut):
    """Thermodynamic integration (TI).

    Parameters
    ----------

    verbose : bool, optional
        Set to True if verbose debug output is desired.

    Attributes
    ----------

    delta_f_ : DataFrame
        The estimated dimensionless free energy difference between each state.

    d_delta_f_ : DataFrame
        The estimated statistical uncertainty (one standard deviation) in
        dimensionless free energy differences.

    states_ : list
        Lambda states for which free energy differences were obtained.

    dhdl : DataFrame
        The estimated dhdl of each state.


    .. versionchanged:: 1.0.0
       `delta_f_`, `d_delta_f_`, `states_` are view of the original object.

    """

    def __init__(self, verbose: bool = False) -> None:
        self.verbose = verbose

    def fit(self, dHdl: pd.DataFrame) -> "TI":
        """
        Compute free energy differences between each state by integrating
        dHdl across lambda values.

        Parameters
        ----------
        dHdl : DataFrame
            dHdl[n,k] is the potential energy gradient with respect to lambda
            for each configuration n and lambda k.

        """

        # sort by state so that rows from same state are in contiguous blocks,
        # and adjacent states are next to each other
        dHdl = dHdl.sort_index(level=dHdl.index.names[1:])  # type: ignore[arg-type]

        # obtain the mean and variance of the mean for each state
        # variance calculation assumes no correlation between points
        # used to calculate mean
        means = dHdl.groupby(level=dHdl.index.names[1:]).mean()
        variances = np.square(dHdl.groupby(level=dHdl.index.names[1:]).sem())

        # get the lambda names
        l_types = dHdl.index.names[1:]

        # obtain vector of delta lambdas between each state
        # Fix issue #148, where for pandas == 1.3.0
        # dl = means.reset_index()[list(means.index.names[:])].diff().iloc[1:].values
        dl = means.reset_index()[means.index.names[:]].diff().iloc[1:].values

        # apply trapezoid rule to obtain DF between each adjacent state
        deltas = (dl * (means.iloc[:-1].values + means.iloc[1:].values) / 2).sum(axis=1)

        # build matrix of deltas between each state
        adelta = np.zeros((len(deltas) + 1, len(deltas) + 1))
        ad_delta = np.zeros_like(adelta)

        for j in range(len(deltas)):
            out = []
            dout = []
            for i in range(len(deltas) - j):
                out.append(deltas[i] + deltas[i + 1 : i + j + 1].sum())

                # Define additional zero lambda
                a = [0.0] * len(l_types)

                # Define dl series' with additional zero lambda on the left and right
                dll = np.insert(dl[i : i + j + 1], 0, [a], axis=0)
                dlr = np.append(dl[i : i + j + 1], [a], axis=0)

                # Get a series of the form: x1, x1 + x2, ..., x(n-1) + x(n), x(n)
                dllr = dll + dlr

                # Append deviation of free energy difference between state i and i+j+1
                dout.append(
                    (dllr**2 * variances.iloc[i : i + j + 2].values / 4)  # type: ignore[attr-defined]
                    .sum(axis=1)
                    .sum()
                )
            adelta += np.diagflat(np.array(out), k=j + 1)
            ad_delta += np.diagflat(np.array(dout), k=j + 1)

        # yield standard delta_f_ free energies between each state
        self._delta_f_ = pd.DataFrame(
            adelta - adelta.T, columns=means.index.values, index=means.index.values
        )
        self.dhdl = means

        # yield standard deviation d_delta_f_ between each state
        self._d_delta_f_ = pd.DataFrame(
            np.sqrt(ad_delta + ad_delta.T),
            columns=variances.index.values,  # type: ignore[attr-defined]
            index=variances.index.values,  # type: ignore[attr-defined]
        )

        self._states_ = means.index.values.tolist()

        self._delta_f_.attrs = dHdl.attrs
        self._d_delta_f_.attrs = dHdl.attrs
        self.dhdl.attrs = dHdl.attrs

        return self

    def separate_dhdl(self) -> list[pd.Series]:
        """
        For transitions with multiple lambda, the attr:`dhdl` would return
        a :class:`~pandas.DataFrame` which gives the dHdl for all the lambda
        states, regardless of whether it is perturbed or not. This function
        creates a list of :class:`pandas.Series` for each lambda, where each
        :class:`pandas.Series` describes the potential energy gradient for the
        lambdas state that is perturbed.

        Returns
        ----------
        dHdl_list : list
            A list of :class:`pandas.Series` such that ``dHdl_list[k]`` is the
            potential energy gradient with respect to lambda for each
            configuration that lambda k is perturbed.
        """
        if len(self.dhdl.index.names) == 1:
            name = self.dhdl.columns[0]
            return [
                self.dhdl[name],
            ]
        dhdl_list = []
        # get the lambda names
        l_types = self.dhdl.index.names
        # obtain bool of changed lambdas between each state
        # Fix issue #148, where for pandas == 1.3.0
        # lambdas = self.dhdl.reset_index()[list(l_types)]
        lambdas = self.dhdl.reset_index()[l_types]
        diff = lambdas.diff().to_numpy(dtype="bool")
        # diff will give the first row as NaN so need to fix that
        diff[0, :] = diff[1, :]
        # Make sure that the start point is set to true as well
        diff[:-1, :] = diff[:-1, :] | diff[1:, :]
        for i in range(len(l_types)):
            if any(diff[:, i]):
                new = self.dhdl.iloc[diff[:, i], i]
                # drop all other index
                for l_type in l_types:
                    if l_type != l_types[i]:
                        new = new.reset_index(l_type, drop=True)
                new.attrs = self.dhdl.attrs
                dhdl_list.append(new)
        return dhdl_list



import pytest

from alchemlyb.parsing.parquet import extract_u_nk, extract_dHdl


@pytest.mark.parametrize(
    "dHdl_list", ["gmx_benzene_Coulomb_dHdl", "gmx_ABFE_complex_dHdl"]
)
def test_extract_dHdl(dHdl_list, request, tmp_path):
    dHdl = request.getfixturevalue(dHdl_list)[0]
    dHdl.to_parquet(path=str(tmp_path / "dhdl.parquet"), index=True)
    new_dHdl = extract_dHdl(str(tmp_path / "dhdl.parquet"), T=300)
    assert (new_dHdl.columns == dHdl.columns).all()
    assert (new_dHdl.index == dHdl.index).all()
    assert new_dHdl.attrs["temperature"] == 300
    assert new_dHdl.attrs["energy_unit"] == "kT"


@pytest.mark.parametrize("u_nk_list", ["gmx_benzene_VDW_u_nk", "gmx_ABFE_complex_u_nk"])
def test_extract_u_nk(u_nk_list, request, tmp_path):
    u_nk = request.getfixturevalue(u_nk_list)[0]
    u_nk.to_parquet(path=str(tmp_path / "u_nk.parquet"), index=True)
    new_u_nk = extract_u_nk(str(tmp_path / "u_nk.parquet"), T=300)
    assert (new_u_nk.columns == u_nk.columns).all()
    assert (new_u_nk.index == u_nk.index).all()
    assert new_u_nk.attrs["temperature"] == 300
    assert new_u_nk.attrs["energy_unit"] == "kT"


@pytest.fixture()
def u_nk(gmx_ABFE_complex_u_nk):
    return gmx_ABFE_complex_u_nk[0]


def test_no_T(u_nk, tmp_path, caplog):
    u_nk.attrs = {}
    u_nk.to_parquet(path=str(tmp_path / "temp.parquet"), index=True)
    extract_u_nk(str(tmp_path / "temp.parquet"), 300)
    assert (
        "Serialise the Dataframe with pandas>=2.1 to preserve the metadata."
        in caplog.text
    )


def test_wrong_T(u_nk, tmp_path, caplog):
    u_nk.to_parquet(path=str(tmp_path / "temp.parquet"), index=True)
    with pytest.raises(ValueError, match="doesn't match the temperature"):
        extract_u_nk(str(tmp_path / "temp.parquet"), 400)


def test_metadata_unchanged(u_nk, tmp_path):
    u_nk.attrs = {"temperature": 400, "energy_unit": "kcal/mol"}
    u_nk.to_parquet(path=str(tmp_path / "temp.parquet"), index=True)
    new_u_nk = extract_u_nk(str(tmp_path / "temp.parquet"), 400)
    assert new_u_nk.attrs["temperature"] == 400
    assert new_u_nk.attrs["energy_unit"] == "kcal/mol"



"""Tests for all TI-based estimators in ``alchemlyb``."""

import pandas as pd
import pytest

import alchemlyb
from alchemlyb.estimators import TI
from alchemlyb.parsing import amber


@pytest.fixture
def Coulomb(gmx_benzene_Coulomb_dHdl):
    dHdl = alchemlyb.concat(gmx_benzene_Coulomb_dHdl)
    return dHdl


@pytest.fixture
def VDW(gmx_benzene_VDW_dHdl):
    dHdl = alchemlyb.concat(gmx_benzene_VDW_dHdl)
    return dHdl


@pytest.fixture
def expanded_ensemble_case_1(gmx_expanded_ensemble_case_1_dHdl):
    dHdl = alchemlyb.concat(gmx_expanded_ensemble_case_1_dHdl)
    return dHdl


@pytest.fixture
def expanded_ensemble_case_2(gmx_expanded_ensemble_case_2_dHdl):
    dHdl = alchemlyb.concat(gmx_expanded_ensemble_case_2_dHdl)
    return dHdl


@pytest.fixture
def expanded_ensemble_case_3(gmx_expanded_ensemble_case_3_dHdl):
    dHdl = alchemlyb.concat(gmx_expanded_ensemble_case_3_dHdl)
    return dHdl


@pytest.fixture
def water_particle_with_total_energy(gmx_water_particle_with_total_energy_dHdl):
    dHdl = alchemlyb.concat(gmx_water_particle_with_total_energy_dHdl)
    return dHdl


@pytest.fixture
def water_particle_with_potential_energy(
    gmx_water_particle_with_potential_energy_dHdl,
):
    dHdl = alchemlyb.concat(gmx_water_particle_with_potential_energy_dHdl)
    return dHdl


@pytest.fixture
def water_particle_without_energy(gmx_water_particle_without_energy_dHdl):
    dHdl = alchemlyb.concat(gmx_water_particle_without_energy_dHdl)
    return dHdl


@pytest.fixture
def simplesolvated_charge(amber_simplesolvated_charge_dHdl):
    dHdl = alchemlyb.concat(amber_simplesolvated_charge_dHdl)
    return dHdl


@pytest.fixture
def simplesolvated_vdw(amber_simplesolvated_vdw_dHdl):
    dHdl = alchemlyb.concat(amber_simplesolvated_vdw_dHdl)
    return dHdl


@pytest.fixture
def benzene(gomc_benzene_dHdl):
    dHdl = alchemlyb.concat(gomc_benzene_dHdl)
    return dHdl


class TIestimatorMixin:
    def test_get_delta_f(self, X_delta_f):
        dHdl, E, dE = X_delta_f
        est = self.cls().fit(dHdl)
        # Use .iloc[0, -1] as we want to cater for both
        # delta_f_.loc[0.0, 1.0] and delta_f_.loc[(0.0, 0.0), (0.0, 1.0)]
        delta_f = est.delta_f_.iloc[0, -1]
        d_delta_f = est.d_delta_f_.iloc[0, -1]

        assert E == pytest.approx(delta_f, rel=1e-3)
        assert dE == pytest.approx(d_delta_f, rel=1e-3)


class TestTI(TIestimatorMixin):
    """Tests for TI."""

    cls = TI

    T = 298.0
    kT_amber = amber.k_b * T

    @pytest.fixture(
        params=[
            ("Coulomb", 3.089, 0.02157),
            ("VDW", -3.056, 0.04863),
            ("expanded_ensemble_case_1", 76.220, 0.15568),
            ("expanded_ensemble_case_2", 76.247, 0.15889),
            ("expanded_ensemble_case_3", 76.387, 0.12532),
            ("water_particle_with_total_energy", -11.696, 0.091775),
            ("water_particle_with_potential_energy", -11.751, 0.091149),
            ("water_particle_without_energy", -11.687, 0.091604),
            ("simplesolvated_charge", -60.114 / kT_amber, 0.08186 / kT_amber),
            ("simplesolvated_vdw", 3.824 / kT_amber, 0.13254 / kT_amber),
        ],
    )
    def X_delta_f(self, request):
        get_dHdl, E, dE = request.param
        return request.getfixturevalue(get_dHdl), E, dE


def test_TI_separate_dhdl_multiple_column(benzene):
    dHdl = benzene
    estimator = TI().fit(dHdl)
    assert all([isinstance(dhdl, pd.Series) for dhdl in estimator.separate_dhdl()])
    assert sorted([len(dhdl) for dhdl in estimator.separate_dhdl()]) == [8, 16]


def test_TI_separate_dhdl_single_column(Coulomb):
    dHdl = Coulomb
    estimator = TI().fit(dHdl)
    assert all([isinstance(dhdl, pd.Series) for dhdl in estimator.separate_dhdl()])
    assert [len(dhdl) for dhdl in estimator.separate_dhdl()] == [
        5,
    ]


def test_TI_separate_dhdl_no_pertubed(Coulomb):
    """The test for the case where two lambda are there and one is not pertubed"""
    dHdl = Coulomb
    dHdl.insert(1, "bound-lambda", [1.0] * len(dHdl))
    dHdl.insert(1, "bound", [1.0] * len(dHdl))
    dHdl.set_index("bound-lambda", append=True, inplace=True)
    estimator = TI().fit(dHdl)
    assert all([isinstance(dhdl, pd.Series) for dhdl in estimator.separate_dhdl()])
    assert [len(dhdl) for dhdl in estimator.separate_dhdl()] == [5]


class Test_Units:
    """Test the units."""

    def test_ti(self, Coulomb):
        ti = TI().fit(Coulomb)
        assert ti.delta_f_.attrs["temperature"] == 300
        assert ti.delta_f_.attrs["energy_unit"] == "kT"
        assert ti.d_delta_f_.attrs["temperature"] == 300
        assert ti.d_delta_f_.attrs["energy_unit"] == "kT"
        assert ti.dhdl.attrs["temperature"] == 300
        assert ti.dhdl.attrs["energy_unit"] == "kT"

    def test_ti_separate_dhdl(self, Coulomb):
        ti = TI().fit(Coulomb)
        dhdl_list = ti.separate_dhdl()
        for dhdl in dhdl_list:
            assert dhdl.attrs["temperature"] == 300
            assert dhdl.attrs["energy_unit"] == "kT"


class Test_MultipleColumnUnits:
    """Test the case where the index has multiple columns"""

    @staticmethod
    @pytest.fixture
    def dhdl(gmx_ABFE_complex_dHdl):
        dhdl = alchemlyb.concat(gmx_ABFE_complex_dHdl)
        return dhdl

    def test_ti_separate_dhdl(self, dhdl):
        ti = TI().fit(dhdl)
        dhdl_list = ti.separate_dhdl()
        for dhdl in dhdl_list:
            assert dhdl.attrs["temperature"] == 300
            assert dhdl.attrs["energy_unit"] == "kT"



from typing import Any
import pandas as pd


class _EstimatorMixOut:
    """This class creates view for the attributes: states_, delta_f_, d_delta_f_,
    delta_h_, d_delta_h_, delta_sT_, d_delta_sT_ for the estimator class to consume.
    """

    _d_delta_f_: pd.DataFrame | None = None
    _delta_f_: pd.DataFrame | None = None
    _states_: list[Any] | None = None
    _d_delta_h_: pd.DataFrame | None = None
    _delta_h_: pd.DataFrame | None = None
    _d_delta_sT_: pd.DataFrame | None = None
    _delta_sT_: pd.DataFrame | None = None

    @property
    def d_delta_f_(self) -> pd.DataFrame | None:
        return self._d_delta_f_

    @property
    def delta_f_(self) -> pd.DataFrame | None:
        return self._delta_f_

    @property
    def d_delta_h_(self) -> pd.DataFrame | None:
        return self._d_delta_h_

    @property
    def delta_h_(self) -> pd.DataFrame | None:
        return self._delta_h_

    @property
    def d_delta_sT_(self) -> pd.DataFrame | None:
        return self._d_delta_sT_

    @property
    def delta_sT_(self) -> pd.DataFrame | None:
        return self._delta_sT_

    @property
    def states_(self) -> list[Any] | None:
        return self._states_



"""Gromacs parser tests."""

import bz2

import pytest
from alchemtest.gmx import load_benzene
from alchemtest.gmx import (
    load_expanded_ensemble_case_1,
    load_expanded_ensemble_case_2,
    load_expanded_ensemble_case_3,
)
from alchemtest.gmx import load_water_particle_with_potential_energy
from alchemtest.gmx import load_water_particle_with_total_energy
from alchemtest.gmx import load_water_particle_without_energy
from numpy.testing import assert_almost_equal

from alchemlyb.parsing.gmx import extract_dHdl, extract_u_nk, extract


def test_dHdl():
    """Test that dHdl has the correct form when extracted from files."""
    dataset = load_benzene()

    for leg in dataset["data"]:
        for filename in dataset["data"][leg]:
            dHdl = extract_dHdl(filename, T=300)

            assert dHdl.index.names == ["time", "fep-lambda"]
            assert dHdl.shape == (4001, 1)


def test_u_nk():
    """Test that u_nk has the correct form when extracted from files."""
    dataset = load_benzene()

    for leg in dataset["data"]:
        for filename in dataset["data"][leg]:
            u_nk = extract_u_nk(filename, T=300)

            assert u_nk.index.names == ["time", "fep-lambda"]
            if leg == "Coulomb":
                assert u_nk.shape == (4001, 5)
            elif leg == "VDW":
                assert u_nk.shape == (4001, 16)


def test_u_nk_case1():
    """Test that u_nk has the correct form when extracted from expanded ensemble files (case 1)."""
    dataset = load_expanded_ensemble_case_1()

    for leg in dataset["data"]:
        for filename in dataset["data"][leg]:
            u_nk = extract_u_nk(filename, T=300, filter=False)

            assert u_nk.index.names == [
                "time",
                "fep-lambda",
                "coul-lambda",
                "vdw-lambda",
                "restraint-lambda",
            ]

            assert u_nk.shape == (50001, 28)


def test_dHdl_case1():
    """Test that dHdl has the correct form when extracted from expanded ensemble files (case 1)."""
    dataset = load_expanded_ensemble_case_1()

    for leg in dataset["data"]:
        for filename in dataset["data"][leg]:
            dHdl = extract_dHdl(filename, T=300, filter=False)

            assert dHdl.index.names == [
                "time",
                "fep-lambda",
                "coul-lambda",
                "vdw-lambda",
                "restraint-lambda",
            ]
            assert dHdl.shape == (50001, 4)


def test_u_nk_case2():
    """Test that u_nk has the correct form when extracted from expanded ensemble files (case 2)."""
    dataset = load_expanded_ensemble_case_2()

    for leg in dataset["data"]:
        for filename in dataset["data"][leg]:
            u_nk = extract_u_nk(filename, T=300, filter=False)

            assert u_nk.index.names == [
                "time",
                "fep-lambda",
                "coul-lambda",
                "vdw-lambda",
                "restraint-lambda",
            ]

            assert u_nk.shape == (25001, 28)


def test_u_nk_case3():
    """Test that u_nk has the correct form when extracted from REX files (case 3)."""
    dataset = load_expanded_ensemble_case_3()

    for leg in dataset["data"]:
        for filename in dataset["data"][leg]:
            u_nk = extract_u_nk(filename, T=300, filter=False)

            assert u_nk.index.names == [
                "time",
                "fep-lambda",
                "coul-lambda",
                "vdw-lambda",
                "restraint-lambda",
            ]

            assert u_nk.shape == (2500, 28)


def test_dHdl_case3():
    """Test that dHdl has the correct form when extracted from REX files (case 3)."""
    dataset = load_expanded_ensemble_case_3()

    for leg in dataset["data"]:
        for filename in dataset["data"][leg]:
            dHdl = extract_dHdl(filename, T=300, filter=False)

            assert dHdl.index.names == [
                "time",
                "fep-lambda",
                "coul-lambda",
                "vdw-lambda",
                "restraint-lambda",
            ]
            assert dHdl.shape == (2500, 4)


def test_u_nk_with_total_energy():
    """Test that the reduced potential is calculated correctly when the total energy is given."""

    # Load dataset
    dataset = load_water_particle_with_total_energy()

    # Check if the sum of values on the diagonal has the correct value
    assert_almost_equal(_diag_sum(dataset), 47611374979.46824, decimal=4)

    # Check one specific value in the dataframe
    assert_almost_equal(
        extract_u_nk(dataset["data"]["AllStates"][0], T=300)
        .loc[0][(0.0, 0.0)]
        .values[0],
        -11211.577658852531,
        decimal=6,
    )


def test_u_nk_with_potential_energy():
    """Test that the reduced potential is calculated correctly when the potential energy is given."""

    # Load dataset
    dataset = load_water_particle_with_potential_energy()

    # Check if the sum of values on the diagonal has the correct value
    assert_almost_equal(_diag_sum(dataset), 16674040406471.553, decimal=2)

    # Check one specific value in the dataframe
    assert_almost_equal(
        extract_u_nk(dataset["data"]["AllStates"][0], T=300)
        .loc[0][(0.0, 0.0)]
        .values[0],
        -15656.557252200757,
        decimal=6,
    )


def test_u_nk_without_energy():
    """Test that the reduced potential is calculated correctly when no energy is given."""

    # Load dataset
    dataset = load_water_particle_without_energy()

    # Check if the sum of values on the diagonal has the correct value
    assert_almost_equal(_diag_sum(dataset), 20572986866779.01, decimal=2)

    # Check one specific value in the dataframe
    assert_almost_equal(
        extract_u_nk(dataset["data"]["AllStates"][0], T=300)
        .loc[0][(0.0, 0.0)]
        .values[0],
        0.0,
        decimal=6,
    )


def _diag_sum(dataset):
    """Calculate the sum of diagonal elements (i, i)"""

    # Initialize the sum variable
    ds = 0.0

    for leg in dataset["data"]:
        for filename in dataset["data"][leg]:
            u_nk = extract_u_nk(filename, T=300)

            # Calculate the sum of diagonal elements:
            for i, lambda_ in enumerate(u_nk.columns):
                # 18.6 is the time step
                ds += u_nk.loc[i * 186 / 10][lambda_].values[0]

    return ds


def test_extract_u_nk_unit():
    """Test if extract_u_nk assign the attr correctly"""
    dataset = load_benzene()
    u_nk = extract_u_nk(dataset["data"]["Coulomb"][0], 310)
    assert u_nk.attrs["temperature"] == 310
    assert u_nk.attrs["energy_unit"] == "kT"


def test_extract_dHdl_unit():
    """Test if extract_u_nk assign the attr correctly"""
    dataset = load_benzene()
    dhdl = extract_dHdl(dataset["data"]["Coulomb"][0], 310)
    assert dhdl.attrs["temperature"] == 310
    assert dhdl.attrs["energy_unit"] == "kT"


def test_calling_extract():
    """Test if the extract function is working"""
    dataset = load_benzene()
    df_dict = extract(dataset["data"]["Coulomb"][0], 310)
    assert df_dict["dHdl"].attrs["temperature"] == 310
    assert df_dict["dHdl"].attrs["energy_unit"] == "kT"
    assert df_dict["u_nk"].attrs["temperature"] == 310
    assert df_dict["u_nk"].attrs["energy_unit"] == "kT"


class TestRobustGMX:
    """Test dropping the row that is wrong in different way"""

    @staticmethod
    @pytest.fixture(scope="class")
    def data():
        dhdl = extract_dHdl(load_benzene()["data"]["Coulomb"][0], 310)
        with bz2.open(load_benzene()["data"]["Coulomb"][0], "rt") as bz_file:
            text = bz_file.read()
        return text, len(dhdl)

    def test_sanity(self, data, tmp_path):
        """Test if the test routine is working."""
        text, length = data
        new_text = tmp_path / "text.xvg"
        new_text.write_text(text)
        dhdl = extract_dHdl(new_text, 310)
        assert len(dhdl) == length

    def test_truncated_row(self, data, tmp_path):
        """Test the case where the last row has been truncated."""
        text, length = data
        new_text = tmp_path / "text.xvg"
        new_text.write_text(text + "40010.0 27.0\n")
        dhdl = extract_dHdl(new_text, 310, filter=True)
        assert len(dhdl) == length

    def test_truncated_number(self, data, tmp_path):
        """Test the case where the last row has been truncated and a - has
        been left."""
        text, length = data
        new_text = tmp_path / "text.xvg"
        new_text.write_text(text + "40010.0 27.0 -\n")
        dhdl = extract_dHdl(new_text, 310, filter=True)
        assert len(dhdl) == length

    def test_weirdnumber(self, data, tmp_path):
        """Test the case where the last number has been appended a weird
        number."""
        text, length = data
        new_text = tmp_path / "text.xvg"
        # Note the 27.040010.0 which is the sum of 27.0 and 40010.0
        new_text.write_text(
            text + "40010.0 27.040010.0 27.0 0.0 6.7 13.5 20.2 27.0 0.7 27.0 0.0 6.7 "
            "13.5 20.2 27.0 0.7\n"
        )
        dhdl = extract_dHdl(new_text, 310, filter=True)
        assert len(dhdl) == length

    def test_too_many_cols(self, data, tmp_path):
        """Test the case where the row has too many columns."""
        text, length = data
        new_text = tmp_path / "text.xvg"
        new_text.write_text(
            text
            + "40010.0 27.0 0.0 6.7 13.5 20.2 27.0 0.7 27.0 0.0 6.7 13.5 20.2 27.0 0.7\n"
        )
        dhdl = extract_dHdl(new_text, 310, filter=True)
        assert len(dhdl) == length



from __future__ import annotations
from typing import Literal

import numpy as np
import pandas as pd
import pymbar
import numpy.typing as npt
from sklearn.base import BaseEstimator

from . import BAR
from .base import _EstimatorMixOut


class MBAR(BaseEstimator, _EstimatorMixOut):
    r"""Multi-state Bennett acceptance ratio (MBAR).

    Parameters
    ----------

    maximum_iterations : int, optional
        Set to limit the maximum number of iterations performed.

    relative_tolerance : float, optional
        Set to determine the relative tolerance convergence criteria.

    initial_f_k : np.ndarray, float, shape=(K), optional or String `BAR`
        When `isinstance(initial_f_k, np.ndarray)`, `initial_f_k` will be used as
        initial guess for MBAR estimator. initial_f_k should be dimensionless
        free energies.
        When `initial_f_k` is ``None``, ``initial_f_k`` will be set to 0.
        When `initial_f_k` is set to "BAR", a BAR calculation will be done and
        the result is used as the initial guess (default).

        .. versionchanged:: 2.3.0
           The new default is now "BAR" as it provides a substantial speedup
           over the previous default `None`.


    method : str, optional, default="robust"
        The optimization routine to use.  This can be any of the methods
        available via :func:`scipy.optimize.minimize` or
        :func:`scipy.optimize.root`.

    n_bootstraps : int, optional
        Whether to use bootstrap to estimate uncertainty. `0` means use analytic error
        estimation. 50~200 is a reasonable range to do bootstrap.

    verbose : bool, optional
        Set to ``True`` if verbose debug output from :mod:`pymbar` is desired.

    Attributes
    ----------

    delta_f_ : DataFrame
        The estimated dimensionless free energy difference between each state.

    d_delta_f_ : DataFrame
        The estimated statistical uncertainty (one standard deviation) in
        dimensionless free energy differences.

    delta_h_ : DataFrame
        The estimated dimensionless enthalpy difference between each state.

    d_delta_h_ : DataFrame
        The estimated statistical uncertainty (one standard deviation) in
        dimensionless enthalpy differences.

    delta_sT_ : DataFrame, optional
        The estimated dimensionless entropy difference between each state.

    d_delta_sT_ : DataFrame
        The estimated statistical uncertainty (one standard deviation) in
        dimensionless entropy differences.

    theta_ : DataFrame
        The theta matrix.

    states_ : list
        Lambda states for which free energy differences were obtained.

    Notes
    -----
    See [Shirts2008]_ for details of the derivation and cite the
    paper when using MBAR in published work.

    See Also
    --------
    pymbar.mbar.MBAR


    .. versionchanged:: 1.0.0
       `delta_f_`, `d_delta_f_`, `states_` are view of the original object.
    .. versionchanged:: 2.0.0
        default value for `method` was changed from "hybr" to "robust"
    .. versionchanged:: 2.1.0
        `n_bootstraps` option added.
    .. versionchanged:: 2.4.0
        Handle initial estimate, initial_f_k, from bar in the instance
        that not all lambda states represented as column headers are
        represented in the indices of u_nk.
    .. versionchanged:: 2.5.0
        Added computation of enthalpy and entropy

    """

    def __init__(
        self,
        maximum_iterations: int = 10000,
        relative_tolerance: float = 1.0e-7,
        initial_f_k: np.ndarray | Literal["BAR"] | None = "BAR",
        method: str = "robust",
        n_bootstraps: int = 0,
        verbose: bool = False,
    ) -> None:
        self.maximum_iterations = maximum_iterations
        self.relative_tolerance = relative_tolerance
        if isinstance(initial_f_k, str) and initial_f_k != "BAR":
            raise ValueError(
                f"Only `BAR` is supported as string input to `initial_f_k`. Got ({initial_f_k})."
            )
        else:
            self.initial_f_k = initial_f_k
        self.method = method
        self.verbose = verbose
        self.n_bootstraps = n_bootstraps

        # handle for pymbar.MBAR object
        self._mbar = None

    def fit(self, u_nk: pd.DataFrame, compute_entropy_enthalpy: bool = False) -> "MBAR":
        """
        Compute overlap matrix of reduced potentials using multi-state
        Bennett acceptance ratio.

        Parameters
        ----------
        u_nk : DataFrame
            ``u_nk[n, k]`` is the reduced potential energy of uncorrelated
            configuration ``n`` evaluated at state ``k``.

        compute_entropy_enthalpy : bool, optional, default=False
            Compute entropy and enthalpy.

        """
        # sort by state so that rows from same state are in contiguous blocks
        u_nk = u_nk.sort_index(level=u_nk.index.names[1:])  # type: ignore[arg-type]

        groups = u_nk.groupby(level=u_nk.index.names[1:])
        N_k = [
            (
                len(groups.get_group(i if isinstance(i, tuple) else (i,)))  # type: ignore[unreachable]
                if i in groups.groups
                else 0
            )
            for i in u_nk.columns
        ]
        self._states_ = u_nk.columns.values.tolist()

        if isinstance(self.initial_f_k, str) and self.initial_f_k == "BAR":
            bar = BAR(
                maximum_iterations=self.maximum_iterations,
                relative_tolerance=self.relative_tolerance,
                verbose=self.verbose,
            )
            bar.fit(u_nk)
            initial_f_k = bar.delta_f_.iloc[0, :]  # type: ignore[union-attr]
            if len(bar.delta_f_.iloc[0, :]) != len(self._states_):  # type: ignore[union-attr]
                states = [
                    x
                    for i, x in enumerate(self._states_[:-1])
                    if N_k[i] > 0 and N_k[i + 1] > 0
                ]
                initial_f_k = pd.Series(
                    [
                        initial_f_k.loc[x] if x in states else np.nan
                        for x in self._states_
                    ],
                    index=self._states_,
                    dtype=float,
                )
        else:
            initial_f_k = self.initial_f_k

        self._mbar = pymbar.MBAR(
            u_nk.T,
            N_k,
            maximum_iterations=self.maximum_iterations,
            relative_tolerance=self.relative_tolerance,
            verbose=self.verbose,
            initial_f_k=initial_f_k,
            solver_protocol=self.method,
            n_bootstraps=self.n_bootstraps,
        )

        uncertainty_method = None if self.n_bootstraps == 0 else "bootstrap"
        out = self._mbar.compute_free_energy_differences(  # type: ignore[attr-defined]
            return_theta=True, uncertainty_method=uncertainty_method
        )
        if compute_entropy_enthalpy:
            out.update(
                self._mbar.compute_entropy_and_enthalpy(  # type: ignore[attr-defined]
                    uncertainty_method=uncertainty_method
                )
            )

        self._delta_f_ = pd.DataFrame(
            out["Delta_f"],
            columns=self._states_,
            index=self._states_,
        )
        self._d_delta_f_ = pd.DataFrame(
            out["dDelta_f"],
            columns=self._states_,
            index=self._states_,
        )
        self.theta_ = pd.DataFrame(
            out["Theta"],
            columns=self._states_,
            index=self._states_,
        )
        if compute_entropy_enthalpy:
            self._delta_h_ = pd.DataFrame(
                out["Delta_u"],
                columns=self._states_,
                index=self._states_,
            )
            self._d_delta_h_ = pd.DataFrame(
                out["dDelta_u"],
                columns=self._states_,
                index=self._states_,
            )
            self._delta_sT_ = pd.DataFrame(
                out["Delta_s"],
                columns=self._states_,
                index=self._states_,
            )
            self._d_delta_sT_ = pd.DataFrame(
                out["dDelta_s"],
                columns=self._states_,
                index=self._states_,
            )

        self._delta_f_.attrs = u_nk.attrs
        self._d_delta_f_.attrs = u_nk.attrs
        if compute_entropy_enthalpy:
            self._delta_h_.attrs = u_nk.attrs  # type: ignore[union-attr]
            self._d_delta_h_.attrs = u_nk.attrs  # type: ignore[union-attr]
            self._delta_sT_.attrs = u_nk.attrs  # type: ignore[union-attr]
            self._d_delta_sT_.attrs = u_nk.attrs  # type: ignore[union-attr]

        return self

    @property
    def overlap_matrix(self) -> npt.NDArray:
        r"""MBAR overlap matrix.

        The estimated state overlap matrix :math:`O_{ij}` is an estimate of the probability
        of observing a sample from state :math:`i` in state :math:`j`.

        The :attr:`overlap_matrix` is computed on-the-fly. Assign it to a variable if
        you plan to re-use it.

        See Also
        ---------
        pymbar.mbar.MBAR.computeOverlap
        """
        return self._mbar.compute_overlap()["matrix"]  # type: ignore[attr-defined, no-any-return]



"""Functions for assessing convergence of free energy estimates and raw data."""

from typing import Any, List, Tuple

import numpy as np
import pandas as pd
from loguru import logger
from sklearn.base import BaseEstimator

from .. import concat
from ..estimators import BAR, TI, MBAR, FEP_ESTIMATORS, TI_ESTIMATORS
from ..postprocessors.units import to_kT

estimators_dispatch = {"BAR": BAR, "TI": TI, "MBAR": MBAR}


def forward_backward_convergence(
    df_list: List[pd.DataFrame],
    estimator: str = "MBAR",
    num: int = 10,
    error_tol: float = 3,
    **kwargs: Any,
) -> pd.DataFrame:
    """Forward and backward convergence of the free energy estimate.

    Generate the free energy estimate as a function of time in both directions,
    with the specified number of equally spaced points in the time
    [Klimovich2015]_. For example, setting `num` to 10 would give the forward
    convergence which is the free energy estimate from the first 10%, 20%, 30%,
    ... of the data. The Backward would give the estimate from the last 10%,
    20%, 30%, ... of the data.

    Parameters
    ----------
    df_list : list
        List of DataFrame of either dHdl or u_nk, where each represents a
        different value of lambda.
    estimator : {'MBAR', 'BAR', 'TI'}
        Name of the estimators.

        .. deprecated:: 1.0.0
           Lower case input is also accepted until release 2.0.0.
    num : int
        The number of blocks used to divide *each* DataFrame and progressively add
        to assess convergence. Note that if the DataFrames are different lengths,
        the number of samples contributed with each block will be different.
    error_tol : float
        The maximum error tolerated for analytic error. If the analytic error is
        bigger than the error tolerance, the bootstrap error will be used.

        .. versionadded:: 2.3.0
        .. versionchanged:: 2.4.0
           Clarified docstring, removed incorrect estimation of std for cumulative
           result in bar and added check that only a single lambda state is
           represented in the indices of each df in df_list.

    kwargs : dict
        Keyword arguments to be passed to the estimator.

    Returns
    -------
    :class:`pandas.DataFrame`
        The DataFrame with convergence data. ::

                Forward  Forward_Error  Backward  Backward_Error  data_fraction
            0  3.016442       0.052748  3.065176        0.051036            0.1
            1  3.078106       0.037170  3.078567        0.036640            0.2
            2  3.072561       0.030186  3.047357        0.029775            0.3
            3  3.048325       0.026070  3.057527        0.025743            0.4
            4  3.049769       0.023359  3.037454        0.023001            0.5
            5  3.034078       0.021260  3.040484        0.021075            0.6
            6  3.043274       0.019642  3.032495        0.019517            0.7
            7  3.035460       0.018340  3.036670        0.018261            0.8
            8  3.042032       0.017319  3.046597        0.017233            0.9
            9  3.044149       0.016405  3.044385        0.016402            1.0


    .. versionadded:: 0.6.0
    .. versionchanged:: 1.0.0
       The ``estimator`` accepts uppercase input.
       The default for using ``estimator='MBAR'`` was changed from
       :class:`~alchemlyb.estimators.MBAR` to :class:`~alchemlyb.estimators.AutoMBAR`.
    .. versionchanged:: 2.0.0
       Use pymbar.MBAR instead of the AutoMBAR option.

    """
    logger.info("Start convergence analysis.")
    logger.info("Check data availability.")

    if estimator not in (FEP_ESTIMATORS + TI_ESTIMATORS):
        msg = f"Estimator {estimator} is not available in {FEP_ESTIMATORS + TI_ESTIMATORS}."
        logger.error(msg)
        raise ValueError(msg)
    else:
        # select estimator class by name
        my_estimator = estimators_dispatch[estimator](**kwargs)
        logger.info(f"Use {estimator} estimator for convergence analysis.")

    # Check that each df in the list has only one value of lambda
    for i, df in enumerate(df_list):
        lambda_values = list(set([x[1:] for x in df.index.to_numpy()]))
        if len(lambda_values) > 1:
            ind = [
                j
                for j in range(len(lambda_values[0]))
                if len(list(set([x[j] for x in lambda_values]))) > 1
            ][0]
            raise ValueError(
                "Provided DataFrame, df_list[{}] has more than one lambda value in df.index[{}]".format(
                    i, ind
                )
            )

    logger.info("Begin forward analysis")
    forward_list = []
    forward_error_list = []
    for i in range(1, num + 1):
        logger.info("Forward analysis: {:.2f}%".format(100 * i / num))
        sample = []
        for data in df_list:
            sample.append(data[: len(data) // num * i])
        mean, error = _forward_backward_convergence_estimate(
            sample, estimator, my_estimator, error_tol, **kwargs
        )
        forward_list.append(mean)
        forward_error_list.append(error)
        logger.info(
            "{:.2f} +/- {:.2f} kT".format(forward_list[-1], forward_error_list[-1])
        )

    logger.info("Begin backward analysis")
    backward_list = []
    backward_error_list = []
    for i in range(1, num + 1):
        logger.info("Backward analysis: {:.2f}%".format(100 * i / num))
        sample = []
        for data in df_list:
            sample.append(data[-len(data) // num * i :])
        mean, error = _forward_backward_convergence_estimate(
            sample, estimator, my_estimator, error_tol, **kwargs
        )
        backward_list.append(mean)
        backward_error_list.append(error)
        logger.info(
            "{:.2f} +/- {:.2f} kT".format(backward_list[-1], backward_error_list[-1])
        )

    convergence = pd.DataFrame(
        {
            "Forward": forward_list,
            "Forward_Error": forward_error_list,
            "Backward": backward_list,
            "Backward_Error": backward_error_list,
            "data_fraction": [i / num for i in range(1, num + 1)],
        }
    )
    convergence.attrs = df_list[0].attrs
    return convergence


def _forward_backward_convergence_estimate(
    sample_list: List[pd.DataFrame],
    estimator: str,
    my_estimator: BaseEstimator,
    error_tol: float,
    **kwargs: Any,
) -> Tuple[float, float]:
    """Use estimator to run the estimation and return the mean and error.

    Parameters
    ----------
    sample_list: A list of samples as pandas Dataframe.
    estimator: The string of the estimator (upper case)
    my_estimator: The estimator object.
    error_tol: The error tolerance.
    kwargs

    Returns
    -------
        mean: The delta_f between 0 and 1
        error: The d_delta_f between 0 and 1
    """
    sample = concat(sample_list)
    result = my_estimator.fit(sample)
    if estimator == "MBAR":
        my_estimator.initial_f_k = result.delta_f_.iloc[0, :]
    mean = result.delta_f_.iloc[0, -1]
    if estimator == "BAR":
        # See https://github.com/alchemistry/alchemlyb/pull/60#issuecomment-430720742
        # Error estimate generated by BAR ARE correlated
        error = np.nan
    else:
        error = result.d_delta_f_.iloc[0, -1]
    if estimator == "MBAR" and error > error_tol:
        logger.warning(
            f"Statistical Error ({error}) bigger than error tolerance ({error_tol}), use bootstrap error instead."
        )
        bootstraps_estimator = estimators_dispatch[estimator](
            n_bootstraps=50, initial_f_k=result.delta_f_.iloc[0, :], **kwargs
        )
        bootstraps_estimator.fit(sample)
        error = bootstraps_estimator.d_delta_f_.iloc[0, -1]

    return mean, error


def _cummean(vals: np.ndarray, out_length: int) -> np.ndarray:
    """The cumulative mean of an array.

    This function computes the cumulative mean and shapes the result to the
    desired length.

    Parameters
    ----------
    vals : numpy.array
        The one-dimensional input array.
    out_length : int
        The length of the output array.

    Returns
    -------
    numpy.array
        The one-dimensional input array with length of total.

    Note
    ----
    If the length of the input series is smaller than the ``out_length``, the
    length of the output array is the same as the input series.


    .. versionadded:: 1.0.0

    """
    in_length = len(vals)
    if in_length < out_length:
        out_length = in_length
    block = in_length // out_length
    reshape = vals[: block * out_length].reshape(block, out_length)
    mean = np.mean(reshape, axis=0)
    result = np.cumsum(mean) / np.arange(1, out_length + 1)
    return np.array(result)


def fwdrev_cumavg_Rc(
    series: pd.Series, precision: float = 0.01, tol: float = 2
) -> Tuple[float, pd.DataFrame]:
    r"""Generate the convergence criteria :math:`R_c` for a single simulation.

    The input will be :class:`pandas.Series` generated by
    :func:`~alchemlyb.preprocessing.subsampling.decorrelate_u_nk` or
    :func:`~alchemlyb.preprocessing.subsampling.decorrelate_dhdl`.

    The output will be the float :math:`R_c` [Fan2020]_ [Fan2021]_ and a
    :class:`pandas.DataFrame` with the forward and backward cumulative average
    at `precision` fractional increments, as described below.

    :math:`R_c = 0` indicates that the system is well equilibrated right from
    the beginning while :math:`R_c = 1` signifies that the whole trajectory is
    not equilibrated.

    Parameters
    ----------
    series : pandas.Series
        The input energy array.
    precision : float
        The precision of the output :math:`R_c`. To speed the calculation up,
        the data has been block-averaged before doing the calculation, the size
        of the block is controlled by the desired precision.
    tol : float
        Tolerance (or convergence threshold :math:`\epsilon` in [Fan2021]_)
        in :math:`kT`.

    Returns
    -------
    float
        Convergence time fraction :math:`R_c` [Fan2021]_
    :class:`pandas.DataFrame`
        The DataFrame with block average. ::

                Forward  Backward  data_fraction
            0  3.016442  3.065176            0.1
            1  3.078106  3.078567            0.2
            2  3.072561  3.047357            0.3
            3  3.048325  3.057527            0.4
            4  3.049769  3.037454            0.5
            5  3.034078  3.040484            0.6
            6  3.043274  3.032495            0.7
            7  3.035460  3.036670            0.8
            8  3.042032  3.046597            0.9
            9  3.044149  3.044385            1.0

    Notes
    -----
    This function computes :math:`R_c` from `equation 16`_ from [Fan2021]_.
    The code is modified based on Shujie Fan's (@VOD555) work. Zhiyi Wu
    (@xiki-tempula) improved the performance of the original algorithm.

    Please cite [Fan2021]_ when using this function.

    See also
    --------
    A_c


    .. versionadded:: 1.0.0

    .. _`equation 16`:
       https://www.ncbi.nlm.nih.gov/pmc/articles/PMC8397498/#FD16

    """
    series = to_kT(series)  # type: ignore[assignment]
    array = series.to_numpy()
    out_length = int(1 / precision)
    g_forward = _cummean(array, out_length)
    g_backward = _cummean(array[::-1], out_length)
    length = len(g_forward)

    convergence = pd.DataFrame(
        {
            "Forward": g_forward,
            "Backward": g_backward,
            "data_fraction": [i / length for i in range(1, length + 1)],
        }
    )
    convergence.attrs = series.attrs

    # Final value
    g = g_forward[-1]
    conv = np.logical_and(np.abs(g_forward - g) < tol, np.abs(g_backward - g) < tol)
    for i in range(out_length):
        if all(conv[i:]):
            return i / length, convergence
    else:
        # This branch exists as we are truncating the dataset to speed the
        # calculation up. For example if the dataset has 21 points and the
        # precision is 0.05. The last point of g_forward is computed using
        # data[0:20] while the last point of g_backward is computed using
        # data[1:21]. Thus, the last point of g_forward and g_backward are not
        # the same as this branch will be triggered.
        return 1.0, convergence


def A_c(series_list: List[pd.Series], precision: float = 0.01, tol: float = 2) -> float:
    r"""Generate the ensemble convergence criteria :math:`A_c` for a set of simulations.

    The input is a :class:`list` of :class:`pandas.Series` generated by
    :func:`~alchemlyb.preprocessing.subsampling.decorrelate_u_nk` or
    :func:`~alchemlyb.preprocessing.subsampling.decorrelate_dhdl`.

    The output will the float :math:`A_c` [Fan2020]_ [Fan2021]_. :math:`A_c` is
    a number between 0 and 1 that can be interpreted as the ratio of the total
    equilibrated simulation time to the whole simulation time for a full set of
    simulations. :math:`A_c = 1` means that all simulation time frames in all
    windows can be considered equilibrated, while :math:`A_c = 0` indicates
    that nothing is equilibrated.

    Parameters
    ----------
    series_list : list
        A list of :class:`pandas.Series` energy array.
    precision : float
        The precision of the output :math:`A_c`. To speed the calculation up, the data
        has been block-averaged before doing the calculation, the size of the
        block is controlled by the desired precision.
    tol : float
        Tolerance (or convergence threshold :math:`\epsilon` in [Fan2021]_)
        in :math:`kT`.

    Returns
    -------
    float
        The area :math:`A_c` under curve for convergence time fraction.

    Notes
    -----
    This function computes :math:`A_c` from `equation 18`_ from [Fan2021]_.

    Please cite [Fan2021]_ when using this function.

    See also
    --------
    fwdrev_cumavg_Rc


    .. versionadded:: 1.0.0

    .. _`equation 18`:
       https://www.ncbi.nlm.nih.gov/pmc/articles/PMC8397498/#FD18

    """
    n_R_c = len(series_list)
    R_c_list = [fwdrev_cumavg_Rc(series, precision, tol)[0] for series in series_list]
    logger.info(f"R_c list: {R_c_list}")
    # Integrate the R_c_list <= R_c over the range of 0 to 1
    array_01 = np.hstack((R_c_list, [0, 1]))
    sorted_array = np.sort(np.unique(array_01))
    result = 0
    for i, element in enumerate(sorted_array[::-1]):
        if i == 0:
            continue
        else:
            d_R_c = sorted_array[-i] - sorted_array[-i - 1]
            result += d_R_c * sum(R_c_list <= element) / n_R_c
    return result


def block_average(
    df_list: List[pd.DataFrame], estimator: str = "MBAR", num: int = 10, **kwargs: Any
) -> pd.DataFrame:
    """Free energy estimate for portions of the trajectory.

    Generate the free energy estimate for a series of blocks in time,
    with the specified number of equally spaced points.
    For example, setting `num` to 10 would give the block averages
    which is the free energy estimate from the first 10% alone, then the
    next 10% ... of the data.

    Parameters
    ----------
    df_list : list
        List of DataFrame of either dHdl or u_nk, where each represents a
        different value of lambda.
    estimator : {'MBAR', 'BAR', 'TI'}
        Name of the estimators.
    num : int
        The number of blocks used to divide *each* DataFrame. Note that
        if the DataFrames are different lengths, the number of samples
        contributed to each block will be different.
    kwargs : dict
        Keyword arguments to be passed to the estimator.

    Returns
    -------
    :class:`pandas.DataFrame`
        The DataFrame with estimate data. ::

               FE             FE_Error
            0  3.016442       0.052748
            1  3.078106       0.037170
            2  3.072561       0.030186
            3  3.048325       0.026070
            4  3.049769       0.023359
            5  3.034078       0.021260
            6  3.043274       0.019642
            7  3.035460       0.018340
            8  3.042032       0.017319
            9  3.044149       0.016405


    .. versionadded:: 2.4.0

    """
    logger.info("Start block averaging analysis.")
    logger.info("Check data availability.")
    if estimator not in (FEP_ESTIMATORS + TI_ESTIMATORS):
        msg = f"Estimator {estimator} is not available in {FEP_ESTIMATORS + TI_ESTIMATORS}."
        logger.error(msg)
        raise ValueError(msg)
    else:
        # select estimator class by name
        estimator_fit = estimators_dispatch[estimator](**kwargs).fit
        logger.info(f"Use {estimator} estimator for convergence analysis.")

    # Check that each df in the list has only one value of lambda
    for i, df in enumerate(df_list):
        lambda_values = list(set([x[1:] for x in df.index.to_numpy()]))
        if len(lambda_values) > 1:
            ind = [
                j
                for j in range(len(lambda_values[0]))
                if len(list(set([x[j] for x in lambda_values]))) > 1
            ][0]
            raise ValueError(
                "Provided DataFrame, df_list[{}] has more than one lambda value in df.index[{}]".format(
                    i, ind
                )
            )

    if estimator in ["BAR"] and len(df_list) > 2:
        raise ValueError(
            "Restrict to two DataFrames, one with a fep-lambda value and one its forward adjacent state for a "
            "meaningful result."
        )

    logger.info("Begin Moving Average Analysis")
    average_list = []
    average_error_list = []
    for i in range(1, num):
        logger.info("Moving Average Analysis: {:.2f}%".format(100 * i / num))
        sample = []
        for data in df_list:
            ind1, ind2 = len(data) // num * (i - 1), len(data) // num * i
            sample.append(data[ind1:ind2])
        sample = concat(sample)  # type: ignore[assignment]
        result = estimator_fit(sample)

        average_list.append(result.delta_f_.iloc[0, -1])
        if estimator.lower() == "bar":
            # See https://github.com/alchemistry/alchemlyb/pull/60#issuecomment-430720742
            # Error estimate generated by BAR ARE correlated
            average_error_list.append(np.nan)
        else:
            average_error_list.append(result.d_delta_f_.iloc[0, -1])
        logger.info(
            "{:.2f} +/- {:.2f} kT".format(average_list[-1], average_error_list[-1])
        )

    convergence = pd.DataFrame(
        {
            "FE": average_list,
            "FE_Error": average_error_list,
        }
    )
    convergence.attrs = df_list[0].attrs
    return convergence



"""NAMD parser tests."""

import bz2
from os.path import basename
from re import search

import pytest
from alchemtest.namd import load_idws
from alchemtest.namd import load_restarted
from alchemtest.namd import load_restarted_reversed
from alchemtest.namd import load_tyr2ala

from alchemlyb.parsing.namd import extract_u_nk, extract


# Indices of lambda values in the following line in NAMD fepout files:
# #NEW FEP WINDOW: LAMBDA SET TO 0.6 LAMBDA2 0.7 LAMBDA_IDWS 0.5
LAMBDA1_IDX_NEW = 6
LAMBDA2_IDX_NEW = 8
LAMBDA_IDWS_IDX_NEW = 10
# Indices of lambda values in the following type of line in NAMD fepout files:
# #Free energy change for lambda window [ 0.6 0.7 ] is 0.12345 ; net change until now is 0.12345
LAMBDA1_IDX_FREE = 7
LAMBDA2_IDX_FREE = 8


@pytest.fixture(scope="module")
def dataset():
    return load_tyr2ala()


@pytest.mark.parametrize(
    "direction,shape",
    [
        ("forward", (21021, 21)),
        ("backward", (21021, 21)),
    ],
)
def test_u_nk(dataset, direction, shape):
    """Test that u_nk has the correct form when extracted from files."""
    for filename in dataset["data"][direction]:
        u_nk = extract_u_nk(filename, T=300)

        assert u_nk.index.names == ["time", "fep-lambda"]
        assert u_nk.shape == shape


def test_u_nk_idws():
    """Test that u_nk has the correct form when extracted from files."""

    filenames = load_idws()["data"]["forward"]
    u_nk = extract_u_nk(filenames, T=300)

    assert u_nk.index.names == ["time", "fep-lambda"]
    assert u_nk.shape == (29252, 11)


@pytest.fixture(params=[load_restarted, load_restarted_reversed])
def restarted_dataset(request):
    return request.param()


def _corrupt_fepout(fepout_in, params, tmp_path):
    """Corrupts specific lines in a fepout file according to each line's prefix,
    using caller-supplied functions.

    Parameters
    ----------
    fepout_in: str
        Path to fepout file to be modified. This file will not be overwritten.

    params: list of tuples
        For each tuple, the first element must be a str that will be passed to
        startswith() to identify the line(s) to modify (e.g. "#NEW"). The
        second element must be a function that accepts a list of strs which is
        the output of running split() on the identified line and returns
        a modified list of tokens that will be reassembled into the line to be
        output.

        The function may return None if this line should not be written. This
        can be used to delete lines from the fepout.

    Returns
    -------
    The name of a temporary file which pytest will unlink.
    """

    fepout_out = tmp_path / basename(fepout_in)
    with bz2.open(fepout_out, "wt") as f_out:
        with bz2.open(fepout_in, "rt") as f_in:
            for line in f_in:
                for prefix, func in params:
                    if line.startswith(prefix):
                        tokens_out = func(line.split())
                        line = (
                            " ".join(tokens_out) + "\n"
                            if tokens_out is not None
                            else None
                        )
                if line is not None:
                    f_out.write(line)
    return str(fepout_out)


@pytest.fixture
def restarted_dataset_inconsistent(restarted_dataset, tmp_path):
    """Returns intentionally messed up dataset where lambda1 and lambda2 at start and end of
    a window are different."""

    filenames = sorted(restarted_dataset["data"]["both"])

    changed = False

    def func_free_line(line_split):
        nonlocal changed
        if float(line_split[7]) >= 0.7 and float(line_split[7]) < 0.9:
            line_split[7] = str(float(line_split[7]) + 0.0001)
            changed = True
        return line_split

    for i in range(len(filenames)):
        filenames[i] = _corrupt_fepout(
            filenames[i], [("#Free", func_free_line)], tmp_path
        )
        # Only actually modify one window so we don't trigger the wrong exception
        if changed is True:
            break

    # Don't directly modify the glob object
    restarted_dataset["data"]["both"] = filenames

    return restarted_dataset


@pytest.fixture
def restarted_dataset_idws_without_lambda_idws(restarted_dataset, tmp_path):
    """Returns intentionally messed up dataset where the first window has IDWS data
    but no lambda_idws.
    """

    # First window won't have any IDWS data so we just drop all its files and fudge the lambdas
    # in the next window to include 0.0 or 1.0 (as appropriate) so we still have a nominally complete calculation

    filenames = [
        x
        for x in sorted(restarted_dataset["data"]["both"])
        if search("000[a-z]?.fepout", x) is None
    ]

    def func_new_line(line_split):
        if float(line_split[LAMBDA1_IDX_NEW]) > 0.5:  # 1->0 (reversed) calculation
            line_split[LAMBDA1_IDX_NEW] == "1.0"
        else:  # regular 0->1 calculation
            line_split[LAMBDA1_IDX_NEW] = "0.0"
        # Drop the lambda_idws
        return line_split[:9]

    def func_free_line(line_split):
        if float(line_split[LAMBDA1_IDX_FREE]) > 0.5:  # 1->0 (reversed) calculation
            line_split[LAMBDA1_IDX_FREE] == "1.0"
        else:  # regular 0->1 calculation
            line_split[LAMBDA1_IDX_FREE] = "0.0"
        return line_split

    filenames[0] = _corrupt_fepout(
        filenames[0], [("#NEW", func_new_line), ("#Free", func_free_line)], tmp_path
    )
    restarted_dataset["data"]["both"] = filenames
    return restarted_dataset


@pytest.fixture
def restarted_dataset_toomany_lambda2(restarted_dataset, tmp_path):
    """Returns intentionally messed up dataset, where there are too many lambda2 values for a
    given lambda1."""

    filenames = sorted(restarted_dataset["data"]["both"])

    # For the same l1 and lidws we retain old lambda2 values thus ensuring a collision
    # Also, don't make a window where lambda1 >= lambda2 because this will trigger the
    # "direction changed" exception instead
    def func_new_line(line_split):
        if float(line_split[LAMBDA2_IDX_NEW]) <= 0.2:
            return line_split
        line_split[LAMBDA1_IDX_NEW] = "0.2"
        if len(line_split) > 9 and line_split[9] == "LAMBDA_IDWS":
            line_split[LAMBDA_IDWS_IDX_NEW] = "0.1"
        return line_split

    def func_free_line(line_split):
        if float(line_split[LAMBDA2_IDX_FREE]) <= 0.2:
            return line_split
        line_split[LAMBDA1_IDX_FREE] = "0.2"
        return line_split

    for i in range(len(filenames)):
        filenames[i] = _corrupt_fepout(
            filenames[i], [("#NEW", func_new_line), ("#Free", func_free_line)], tmp_path
        )

    restarted_dataset["data"]["both"] = filenames
    return restarted_dataset


@pytest.fixture
def restarted_dataset_toomany_lambda_idws(restarted_dataset, tmp_path):
    """Returns intentionally messed up dataset, where there are too many lambda2 values for a
    given lambda1."""

    filenames = sorted(restarted_dataset["data"]["both"])

    # For the same lambda1 and lambda2 we retain the first set of lambda1/lambda2 values
    # and replicate them across all windows thus ensuring that there will be more than
    # one lambda_idws value for a given lambda1 and lambda2
    this_lambda1, this_lambda2 = None, None

    def func_new_line(line_split):
        nonlocal this_lambda1, this_lambda2

        if this_lambda1 is None:
            this_lambda1, this_lambda2 = (
                line_split[LAMBDA1_IDX_NEW],
                line_split[LAMBDA2_IDX_NEW],
            )
        # Ensure that changing these lambda values won't cause a reversal in direction and trigger
        # an exception we're not trying to test here
        if len(line_split) > 9 and float(line_split[LAMBDA_IDWS_IDX_NEW]) < 0.5:
            line_split[LAMBDA1_IDX_NEW], line_split[LAMBDA2_IDX_NEW] = (
                this_lambda1,
                this_lambda2,
            )
        return line_split

    def func_free_line(line_split):
        line_split[LAMBDA1_IDX_FREE], line_split[LAMBDA2_IDX_FREE] = (
            this_lambda1,
            this_lambda2,
        )
        return line_split

    for i in range(len(filenames)):
        filenames[i] = _corrupt_fepout(
            filenames[i], [("#NEW", func_new_line)], tmp_path
        )

    restarted_dataset["data"]["both"] = filenames
    return restarted_dataset


@pytest.fixture
def restarted_dataset_direction_changed(restarted_dataset, tmp_path):
    """Returns intentionally messed up dataset, with one window where the lambda values are reversed."""

    filenames = sorted(restarted_dataset["data"]["both"])

    def func_new_line(line_split):
        line_split[6], line_split[8], line_split[10] = (
            line_split[10],
            line_split[8],
            line_split[6],
        )
        return line_split

    def func_free_line(line_split):
        line_split[7], line_split[8] = line_split[8], line_split[7]
        return line_split

    # Reverse the direction of lambdas for this window
    idx_to_corrupt = filenames.index(sorted(filenames)[-3])
    fname1 = _corrupt_fepout(
        filenames[idx_to_corrupt],
        [("#NEW", func_new_line), ("#Free", func_free_line)],
        tmp_path,
    )
    filenames[idx_to_corrupt] = fname1
    restarted_dataset["data"]["both"] = filenames
    return restarted_dataset


@pytest.fixture
def restarted_dataset_all_windows_truncated(restarted_dataset, tmp_path):
    """Returns dataset where all windows are truncated (no #Free... footer lines)."""

    filenames = sorted(restarted_dataset["data"]["both"])

    def func_free_line(line_split):
        return None

    for i in range(len(filenames)):
        filenames[i] = _corrupt_fepout(
            filenames[i], [("#Free", func_free_line)], tmp_path
        )

    restarted_dataset["data"]["both"] = filenames
    return restarted_dataset


@pytest.fixture
def restarted_dataset_last_window_truncated(restarted_dataset, tmp_path):
    """Returns dataset where the last window is truncated (no #Free... footer line)."""

    filenames = sorted(restarted_dataset["data"]["both"])

    def func_free_line(line_split):
        return None

    filenames[-1] = _corrupt_fepout(
        filenames[-1], [("#Free", func_free_line)], tmp_path
    )
    restarted_dataset["data"]["both"] = filenames
    return restarted_dataset


def test_u_nk_restarted():
    """Test that u_nk has the correct form when extracted from an IDWS
    FEP run that includes terminations and restarts.
    """
    filenames = load_restarted()["data"]["both"]
    u_nk = extract_u_nk(filenames, T=300)

    assert u_nk.index.names == ["time", "fep-lambda"]
    assert u_nk.shape == (30061, 11)


def test_u_nk_restarted_missing_window_header(tmp_path):
    """Test that u_nk has the correct form when a #NEW line is missing from the restarted dataset
    and the parser has to infer lambda_idws for that window."""
    filenames = sorted(load_restarted()["data"]["both"])
    # Remove "#NEW" line
    filenames[4] = _corrupt_fepout(
        filenames[4],
        [
            ("#NEW", lambda line_split: None),
        ],
        tmp_path,
    )

    u_nk = extract_u_nk(filenames, T=300)

    assert u_nk.index.names == ["time", "fep-lambda"]
    assert u_nk.shape == (30061, 11)


def test_u_nk_restarted_reversed():
    filenames = load_restarted_reversed()["data"]["both"]
    u_nk = extract_u_nk(filenames, T=300)

    assert u_nk.index.names == ["time", "fep-lambda"]
    assert u_nk.shape == (30170, 11)


def test_extract():
    filenames = load_restarted_reversed()["data"]["both"]
    df_dict = extract(filenames, T=300)

    assert df_dict["u_nk"].index.names == ["time", "fep-lambda"]
    assert df_dict["u_nk"].shape == (30170, 11)
    assert "dHdl" not in df_dict


def test_u_nk_restarted_reversed_missing_window_header(tmp_path):
    """Test that u_nk has the correct form when a #NEW line is missing from the restarted_reversed dataset
    and the parser has to infer lambda_idws for that window."""
    filenames = sorted(load_restarted_reversed()["data"]["both"])
    # Remove "#NEW" line
    filenames[4] = _corrupt_fepout(
        filenames[4],
        [
            ("#NEW", lambda line_split: None),
        ],
        tmp_path,
    )

    u_nk = extract_u_nk(filenames, T=300)

    assert u_nk.index.names == ["time", "fep-lambda"]
    assert u_nk.shape == (30170, 11)


def test_u_nk_restarted_direction_changed(restarted_dataset_direction_changed):
    """Test that when lambda values change direction within a dataset, parsing throws an error."""

    with pytest.raises(ValueError, match="Lambda values change direction"):
        extract_u_nk(restarted_dataset_direction_changed["data"]["both"], T=300)


def test_u_nk_restarted_idws_without_lambda_idws(
    restarted_dataset_idws_without_lambda_idws,
):
    """Test that when the first window has IDWS data but no lambda_idws, parsing throws an error.

    In this situation, the lambda_idws cannot be inferred, because there's no previous lambda
    value available.
    """

    with pytest.raises(
        ValueError,
        match="IDWS data present in first window but lambda_idws not included",
    ):
        extract_u_nk(restarted_dataset_idws_without_lambda_idws["data"]["both"], T=300)


def test_u_nk_restarted_inconsistent(restarted_dataset_inconsistent):
    """Test that when lambda values are inconsistent between start and end of a single window,
    parsing throws an error.
    """

    with pytest.raises(
        ValueError, match="Inconsistent lambda values within the same window"
    ):
        extract_u_nk(restarted_dataset_inconsistent["data"]["both"], T=300)


def test_u_nk_restarted_toomany_lambda_idws(restarted_dataset_toomany_lambda_idws):
    """Test that when there is more than one lambda_idws for a given lambda1, parsing throws an error."""

    with pytest.raises(
        ValueError, match="More than one lambda_idws value for a particular lambda1"
    ):
        extract_u_nk(restarted_dataset_toomany_lambda_idws["data"]["both"], T=300)


def test_u_nk_restarted_toomany_lambda2(restarted_dataset_toomany_lambda2):
    """Test that when there is more than one lambda2 for a given lambda1, parsing throws an error."""

    with pytest.raises(
        ValueError, match="More than one lambda2 value for a particular lambda1"
    ):
        extract_u_nk(restarted_dataset_toomany_lambda2["data"]["both"], T=300)


def test_u_nk_restarted_all_windows_truncated(restarted_dataset_all_windows_truncated):
    """Test that when there is more than one lambda2 for a given lambda1, parsing throws an error."""

    with pytest.raises(ValueError, match="New window begun after truncated window"):
        extract_u_nk(restarted_dataset_all_windows_truncated["data"]["both"], T=300)


def test_u_nk_restarted_last_window_truncated(restarted_dataset_last_window_truncated):
    """Test that when there is more than one lambda2 for a given lambda1, parsing throws an error."""

    with pytest.raises(ValueError, match="Last window is truncated"):
        extract_u_nk(restarted_dataset_last_window_truncated["data"]["both"], T=300)



"""Parsers for extracting alchemical data from `Gromacs <http://www.gromacs.org/>`_ output files."""

import numpy as np
import pandas as pd
from typing import Any

from . import _init_attrs
from .util import anyopen
from ..postprocessors.units import R_kJmol

k_b = R_kJmol


@_init_attrs
def extract_u_nk(xvg: str, T: float, filter: bool = True) -> pd.DataFrame:
    r"""Return reduced potentials `u_nk` from a Hamiltonian differences XVG file.

    Parameters
    ----------
    xvg : str
        Path to XVG file to extract data from.
    T : float
        Temperature in Kelvin the simulations sampled.
    filter : bool
        Filter out the lines that cannot be parsed.
        Such as rows with incorrect number of Columns and incorrectly
        formatted numbers (e.g. 123.45.67, nan or -).

    Returns
    -------
    u_nk : DataFrame
        Potential energy for each alchemical state (k) for each frame (n).


    Note
    -----
    Previous versions of alchemlyb (<0.5.0) used the `GROMACS value of the
    molar gas constant
    <https://manual.gromacs.org/documentation/2019/reference-manual/definitions.html>`_
    of :math:`R = 8.3144621 \times 10^{−3}\,
    \text{kJ}\cdot\text{mol}^{-1}\cdot\text{K}^{-1}` instead of the scipy value
    :data:`scipy.constants.R` in :mod:`scipy.constants` (see
    :mod:`alchemlyb.postprocessors.units`).  The relative difference between
    the two values is :math:`6 \times 10^{-8}`.

    Therefore, results in :math:`kT` for GROMACS data will differ between
    alchemlyb ≥0.5.0 and previous versions; the relative difference is on the
    order of :math:`10^{-7}` for typical cases.


    .. versionchanged:: 0.5.0
        The :mod:`scipy.constants` is used for parsers instead of
        the constants used by the corresponding MD engine.
        This leads to slightly different results for GROMACS input compared to
        previous versions of alchemlyb.

    .. versionchanged:: 0.7.0
        The keyword filter is implemented to ignore the line that cannot be
        parsed and is turned on by default.

    """

    h_col_match = r"\xD\f{}H \xl\f{}"
    pv_col_match = "pV"
    u_col_match = ["Total Energy", "Potential Energy"]
    beta = 1 / (k_b * T)

    state, lambdas, statevec = _extract_state(xvg)

    # extract a DataFrame from XVG data
    df = _extract_dataframe(xvg, filter=filter)

    times = df[df.columns[0]]

    # want to grab only dH columns
    DHcols = [col for col in df.columns if (h_col_match in col)]
    dH = df[DHcols]

    # gromacs also gives us pV directly; need this for reduced potential
    pv_cols = [col for col in df.columns if (pv_col_match in col)]
    pv = None
    if pv_cols:
        pv = df[pv_cols[0]]

    # gromacs also gives us total/potential energy U directly; need this for reduced potential
    u_cols = [
        col
        for col in df.columns
        if any(single_u_col_match in col for single_u_col_match in u_col_match)
    ]
    u = None
    if u_cols:
        u = df[u_cols[0]]

    u_k = dict()
    cols = list()
    for col in dH:
        u_col = eval(col.split("to")[1])  # type: ignore[attr-defined]
        # calculate reduced potential u_k = dH + pV + U
        u_k[u_col] = beta * dH[col].values
        if pv_cols:
            u_k[u_col] += beta * pv.values  # type: ignore[union-attr]
        if u_cols:
            u_k[u_col] += beta * u.values  # type: ignore[union-attr]
        cols.append(u_col)

    u_k = pd.DataFrame(
        u_k, columns=cols, index=pd.Index(times.values, name="time", dtype="Float64")
    )  # type: ignore[assignment]

    # create columns for each lambda, indicating state each row sampled from
    # if state is None run as expanded ensemble data or REX
    if state is None:
        # if thermodynamic state is specified map thermodynamic
        # state data to lambda values, else (for REX)
        # define state based on the legend
        if "Thermodynamic state" in df:
            ts_index = df.columns.get_loc("Thermodynamic state")
            thermo_state = df[df.columns[ts_index]]
            for i, lambda_value in enumerate(lambdas):
                v = []
                for t in thermo_state:
                    v.append(statevec[int(t)][i])  # type: ignore[index, call-overload]
                u_k[lambda_value] = v
        else:
            state_legend = _extract_legend(xvg)
            for i, state_legend_item in enumerate(state_legend):
                u_k[state_legend_item] = state_legend[state_legend_item]
    else:
        for i, lambda_value in enumerate(lambdas):
            try:
                u_k[lambda_value] = statevec[i]
            except TypeError:
                u_k[lambda_value] = statevec

    # set up new multi-index
    newind = ["time"] + lambdas
    u_k = u_k.reset_index().set_index(newind)  # type: ignore[attr-defined]

    u_k.name = "u_nk"

    return u_k  # type: ignore[no-any-return]


@_init_attrs
def extract_dHdl(xvg: str, T: float, filter: bool = True) -> pd.DataFrame:
    r"""Return gradients `dH/dl` from a Hamiltonian differences XVG file.

    Parameters
    ----------
    xvg : str
        Path to XVG file to extract data from.
    T : float
        Temperature in Kelvin the simulations sampled.
    filter : bool
        Filter out the lines that cannot be parsed.
        Such as rows with incorrect number of Columns and incorrectly
        formatted numbers (e.g. 123.45.67, nan or -).

    Returns
    -------
    dH/dl : Series
        dH/dl as a function of time for this lambda window.

    Note
    -----
    Previous versions of alchemlyb (<0.5.0) used the `GROMACS value of the
    molar gas constant
    <https://manual.gromacs.org/documentation/2019/reference-manual/definitions.html>`_
    of :math:`R = 8.3144621 \times 10^{−3}\,
    \text{kJ}\cdot\text{mol}^{-1}\cdot\text{K}^{-1}` instead of the scipy value
    :data:`scipy.constants.R` in :mod:`scipy.constants` (see
    :mod:`alchemlyb.postprocessors.units`).  The relative difference between
    the two values is :math:`6 \times 10^{-8}`.

    Therefore, results in :math:`kT` for GROMACS data will differ between
    alchemlyb ≥0.5.0 and previous versions; the relative difference is on the
    order of :math:`10^{-7}` for typical cases.


    .. versionchanged:: 0.5.0
        The :mod:`scipy.constants` is used for parsers instead of
        the constants used by the corresponding MD engine.
        This leads to slightly different results for GROMACS input compared to
        previous versions of alchemlyb.

    .. versionchanged:: 0.7.0
        The keyword filter is implemented to ignore the line that cannot be
        parsed and is turned on by default.

    """
    beta = 1 / (k_b * T)

    headers = _get_headers(xvg)
    state, lambdas, statevec = _extract_state(xvg, headers)

    # extract a DataFrame from XVG data
    df = _extract_dataframe(xvg, headers, filter=filter)

    times = df[df.columns[0]]

    # want to grab only dH/dl columns
    dHcols = []
    for lambda_value in lambdas:
        dHcols.extend([col for col in df.columns if (lambda_value in col)])

    dHdl = df[dHcols]

    # make dimensionless
    dHdl = beta * dHdl

    # rename columns to not include the word 'lambda', since we use this for
    # index below
    cols = [lambda_value.split("-")[0] for lambda_value in lambdas]

    dHdl = pd.DataFrame(
        dHdl.values,
        columns=cols,
        index=pd.Index(times.values, name="time", dtype="Float64"),
    )

    # create columns for each lambda, indicating state each row sampled from
    # if state is None run as expanded ensemble data or REX
    if state is None:
        # if thermodynamic state is specified map thermodynamic
        # state data to lambda values, else (for REX)
        # define state based on the legend
        if "Thermodynamic state" in df:
            ts_index = df.columns.get_loc("Thermodynamic state")
            thermo_state = df[df.columns[ts_index]]
            for i, lambda_value in enumerate(lambdas):
                v = []
                for t in thermo_state:
                    v.append(statevec[int(t)][i])  # type: ignore[index, call-overload]
                dHdl[lambda_value] = v
        else:
            state_legend = _extract_legend(xvg)
            for i, state_legend_item in enumerate(state_legend):
                dHdl[state_legend_item] = state_legend[state_legend_item]
    else:
        for i, lambda_value in enumerate(lambdas):
            try:
                dHdl[lambda_value] = statevec[i]
            except TypeError:
                dHdl[lambda_value] = statevec

    # set up new multi-index
    newind = ["time"] + lambdas
    dHdl = dHdl.reset_index().set_index(newind)

    dHdl.name = "dH/dl"

    return dHdl


def extract(xvg: str, T: float, filter: bool = True) -> dict[str, pd.DataFrame | None]:
    r"""Return reduced potentials `u_nk` and gradients `dH/dl`
    from a Hamiltonian differences XVG file.

    Parameters
    ----------
    xvg : str
        Path to XVG file to extract data from.
    T : float
        Temperature in Kelvin the simulations sampled.
    filter : bool
        Filter out the lines that cannot be parsed.
        Such as rows with incorrect number of Columns and incorrectly
        formatted numbers (e.g. 123.45.67, nan or -).

    Returns
    -------
    Dict
        A dictionary with keys of 'u_nk', which is a pandas DataFrame of
        potential energy for each alchemical state (k) for each frame (n),
        and 'dHdl', which is a Series of dH/dl
        as a function of time for this lambda window.


    .. versionadded:: 1.0.0
    """

    return {"u_nk": extract_u_nk(xvg, T, filter), "dHdl": extract_dHdl(xvg, T, filter)}


def _extract_state(
    xvg: str, headers: None | dict[str, Any] = None
) -> tuple[None | int, list[str], list[float]]:
    """Extract information on state sampled, names of lambdas.

    Parameters
    ----------
    xvg : str
        Path to XVG file to extract data from.
    headers: dict
       headers dictionary to search header information, reduced I/O by
       reusing if it is already parsed, e.g. _extract_state and
       _extract_dataframe in order need one-time header parsing

    """
    state = None
    if headers is None:
        headers = _get_headers(xvg)
    subtitle = _get_value_by_key(headers, "subtitle")
    if subtitle and "state" in subtitle:
        state = int(subtitle.split("state")[1].split(":")[0])
        lambdas = [word.strip(")(,") for word in subtitle.split() if "lambda" in word]
        statevec = eval(subtitle.strip().split(" = ")[-1].strip('"'))

    # if expanded ensemble data is used the state variable will never be assigned
    # parsing expanded ensemble data
    if state is None:
        lambdas = []
        statevec = []
        for line in headers["_raw_lines"]:
            if ("legend" in line) and ("lambda" in line):
                lambdas.append(
                    [word.strip(")(,") for word in line.split() if "lambda" in word][0]
                )
            if ("legend" in line) and (" to " in line):
                statevec.append(
                    (
                        [
                            float(i)
                            for i in line.strip()
                            .split(" to ")[-1]
                            .strip('"()')
                            .split(",")
                        ]
                    )
                )

    return state, lambdas, statevec


def _extract_legend(xvg: str) -> dict[str, float]:
    """Extract information on state sampled for REX simulations."""
    state_legend = {}
    with anyopen(xvg, "r") as f:  # type: ignore[arg-type]
        for line in f:
            if ("legend" in line) and ("lambda" in line):
                state_legend[line.split()[4]] = float(line.split()[6].strip('"'))

    return state_legend


def _extract_dataframe(
    xvg: str, headers: None | dict[str, Any] = None, filter: bool = True
) -> pd.DataFrame:
    """Extract a DataFrame from XVG data using Pandas `read_csv()`.

    pd.read_csv() shows the same behavior building pandas Dataframe with better
    performance (approx. 2 to 4 times speed up). See Issue #81.

    Parameters
    ----------
    xvg: str
       Path to XVG file to extract data from.
    headers: dict
       headers dictionary to search header information, reduced I/O by
       reusing if it is already parsed. Direct access by key name
    filter : bool
        Filter out the lines that cannot be parsed.
        Such as rows with incorrect number of Columns and incorrectly
        formatted numbers (e.g. 123.45.67, nan or -).

    """
    if headers is None:
        headers = _get_headers(xvg)

    xaxis = _get_value_by_key(headers, "xaxis", "label")
    names = [
        _get_value_by_key(headers, "s{}".format(x), "legend")
        for x in range(len(headers))
        if "s{}".format(x) in headers
    ]
    cols = [xaxis] + names

    # march through column names, mark duplicates when found
    cols = [
        col + "{}[duplicated]".format(i) if col in cols[:i] else col  # type: ignore[operator]
        for i, col in enumerate(cols)
    ]

    header_cnt = len(headers["_raw_lines"])
    if not filter:
        # assumes clean files
        df = pd.read_csv(
            xvg,
            sep=r"\s+",
            header=None,
            skiprows=header_cnt,
            na_filter=True,
            memory_map=True,
            names=cols,
            dtype=np.float64,
            float_precision="high",
        )
    else:
        df = pd.read_csv(
            xvg,
            sep=r"\s+",
            header=None,
            skiprows=header_cnt,
            memory_map=True,
            on_bad_lines="skip",
        )
        # If names=cols is passed to read_csv, rows with more than the
        # designated columns will be truncated and used instead of discarded.
        df.rename(columns={i: name for i, name in enumerate(cols)}, inplace=True)
        # If dtype=np.float64 and float_precision='high' are passed to read_csv,
        # 12.345.56 and - cannot be read.
        df = df.apply(pd.to_numeric, errors="coerce")
        # drop duplicate
        df.dropna(inplace=True)

    # drop duplicated columns (see PR #86 https://github.com/alchemistry/alchemlyb/pull/86/)
    df = df[df.columns[~df.columns.str.endswith("[duplicated]")]]

    return df


def _parse_header(line: str, headers: dict[str, Any] = {}, depth: int = 2) -> None:
    """Build python dictionary for single line header

    Update python dictionary to ``headers`` by reading ``line`` separated by
    whitespace. If ``depth`` is given, at most ``depth`` nested key value store
    is added. `_val` key is reserved which contain remaining words from
    ``line``.

    Note
    ----
    No return value but 'headers' dictionary will be updated.

    Parameters
    ----------

    line: str
        header line to parse
    headers: dict
        headers dictionary to update, pass by reference
    depth: int
        depth of nested key and value store

    Examples
    --------
    "x y z" line turns into { 'x': { 'y': {'_val': 'z' }}}

    >>> headers={}
    >>> _parse_header('@ s0 legend "Potential Energy (kJ/mol)"', headers)
    >>> headers
    {'s0': {'legend': {'_val': 'Potential Energy (kJ/mol)'}}}

    """
    # Remove a first character, i.e. @
    s = line[1:].split(None, 1)
    next_t = headers[s[0]] = {}
    for i in range(1, depth):
        # ord('"') == 34
        # no further parsing for quoted value
        if len(s) > 1 and s[1][0] != '"':
            s = s[1].split(None, 1)
            next_t[s[0]] = {}
            next_t = next_t[s[0]]
        else:
            break

    next_t["_val"] = "".join(s[1:]).rstrip().strip('"')


def _get_headers(xvg: str) -> dict[str, Any]:
    """Build python dictionary from header lines

    Build nested key and value store by reading header ('@') lines from a file.
    Direct access to value provides reduced time complexity O(1).
    `_raw_lines` key is reserved to keep the original text.

    Example
    -------

    Given a xvg file containinig header lines like:

        ...
       @    title "dH/d\\xl\\f{} and \\xD\\f{}H"
       @    xaxis  label "Time (ps)"
       @    yaxis  label "dH/d\\xl\\f{} and \\xD\\f{}H (kJ/mol [\\xl\\f{}]\\S-1\\N)"
       @TYPE xy
       @ subtitle "T = 310 (K) \\xl\\f{} state 38: (coul-lambda, vdw-lambda) = (0.9500, 0.0000)"
       @ view 0.15, 0.15, 0.75, 0.85
       @ legend on
       @ legend box on
       @ legend loctype view
       @ legend 0.78, 0.8
       @ legend length 2
       @ s0 legend "Potential Energy (kJ/mol)"
       @ s1 legend "dH/d\\xl\\f{} coul-lambda = 0.9500"
       @ s2 legend "dH/d\\xl\\f{} vdw-lambda = 0.0000"
       ...

    >>> _get_headers(xvg)
    {'TYPE': {'xy': {'_val': ''}},
      'subtitle': {'_val': 'T = 310 (K) \\xl\\f{} state 38: (coul-lambda, vdw-lambda) = (0.9500, 0.0000)'},
      'title': {'_val': 'dH/d\\xl\\f{} and \\xD\\f{}H'},
      'view': {'0.15,': {'_val': '0.15, 0.75, 0.85'}},
      'xaxis': {'label': {'_val': 'Time (ps)'}},
      'yaxis': {'label': {'_val': 'dH/d\\xl\\f{} and \\xD\\f{}H (kJ/mol [\\xl\\f{}]\\S-1\\N)'}},
      ...(omitted)...
      '_raw_lines': ['@    title "dH/d\\xl\\f{} and \\xD\\f{}H"',
                    '@    xaxis  label "Time (ps)"',
                    '@    yaxis  label "dH/d\\xl\\f{} and \\xD\\f{}H (kJ/mol [\\xl\\f{}]\\S-1\\N)"',
                    '@TYPE xy',
                    '@ subtitle "T = 310 (K) \\xl\\f{} state 38: (coul-lambda, vdw-lambda) = (0.9500, 0.0000)"',
                    '@ view 0.15, 0.15, 0.75, 0.85',
                    '@ legend on',
                    '@ legend box on',
                    '@ legend loctype view',
                    '@ legend 0.78, 0.8',
                    '@ legend length 2',
                    '@ s0 legend "Potential Energy (kJ/mol)"',
                    '@ s1 legend "dH/d\\xl\\f{} coul-lambda = 0.9500"',
                    '@ s2 legend "dH/d\\xl\\f{} vdw-lambda = 0.0000"'],
      }

    Returns
    -------
    headers: dict

    """
    with anyopen(xvg, "r") as f:  # type: ignore[arg-type]
        headers: dict[str, Any] = {"_raw_lines": []}
        for line in f:
            line = line.strip()
            if len(line) == 0:
                continue
            if line.startswith("@"):
                _parse_header(line, headers)
                headers["_raw_lines"].append(line)
            elif line.startswith("#"):
                headers["_raw_lines"].append(line)
                continue
            # assuming to start a body section
            else:
                break

    return headers


def _get_value_by_key(
    headers: dict[str, Any], key1: str, key2: None | str = None
) -> None | str:
    """Return value by two-level keys where the second key is optional

    Example
    -------

    >>> headers
    {'s0': {'legend': {'_val': 'Potential Energy (kJ/mol)'}},
            'subtitle': {'_val': 'T = 310 (K) \\xl\\f{} state 38: (coul-lambda,
                vdw-lambda) = (0.9500, 0.0000)'}}
    >>> _get_value_by_key(header, 's0','legend')
    'Potential Energy (kJ/mol)'
    >>> _get_value_by_key(header, 'subtitle')
    'T = 310 (K) \\xl\\f{} state 38: (coul-lambda, vdw-lambda) = (0.9500, 0.0000)'

    """
    val = None
    if key1 in headers:
        if key2 is not None and key2 in headers[key1]:
            val = headers[key1][key2]["_val"]
        else:
            val = headers[key1]["_val"]

    return val



from .bar_ import BAR
from .mbar_ import MBAR
from .ti_ import TI
from .ti_gaussian_quadrature_ import TI_GQ

FEP_ESTIMATORS = [MBAR.__name__, BAR.__name__]
TI_ESTIMATORS = [TI.__name__, TI_GQ.__name__]



"""Tests for all FEP-based estimators in ``alchemlyb``."""

import numpy as np
import pytest

import alchemlyb
from alchemlyb.estimators import MBAR, BAR


class FEPestimatorMixin:
    """Mixin for all FEP Estimator test classes."""

    def compare_delta_f(self, X_delta_f):
        est = self.cls().fit(X_delta_f[0])
        delta_f, d_delta_f = self.get_delta_f(est)

        assert X_delta_f[1] == pytest.approx(delta_f, rel=1e-3)
        assert X_delta_f[2] == pytest.approx(d_delta_f, rel=1e-3)

    def get_delta_f(self, est):
        # Use .iloc[0, -1] as we want to cater for both
        # delta_f_.loc[0.0, 1.0] and delta_f_.loc[(0.0, 0.0), (0.0, 1.0)]
        return est.delta_f_.iloc[0, -1], est.d_delta_f_.iloc[0, -1]


class TestMBAR(FEPestimatorMixin):
    """Tests for MBAR."""

    cls = MBAR

    @pytest.fixture(
        params=[
            ("gmx_benzene_Coulomb_u_nk", 3.041, 0.02088),
            ("gmx_benzene_VDW_u_nk", -3.007, 0.04519),
            ("gmx_expanded_ensemble_case_1", 75.923, 0.14124),
            ("gmx_expanded_ensemble_case_2", 75.915, 0.14372),
            ("gmx_expanded_ensemble_case_3", 76.173, 0.11345),
            ("gmx_water_particle_with_total_energy", -11.680, 0.083655),
            ("gmx_water_particle_with_potential_energy", -11.675, 0.083589),
            ("gmx_water_particle_without_energy", -11.654, 0.083415),
            ("amber_bace_example_complex_vdw", 2.41149, 0.0620658),
            ("gomc_benzene_u_nk", -0.79994, 0.091579),
        ],
    )
    def X_delta_f(self, request):
        get_unk, E, dE = request.param
        return alchemlyb.concat(request.getfixturevalue(get_unk)), E, dE

    def test_mbar(self, X_delta_f):
        self.compare_delta_f(X_delta_f)


class TestBAR(FEPestimatorMixin):
    """Tests for BAR."""

    cls = BAR

    @pytest.fixture(
        params=[
            ("gmx_benzene_Coulomb_u_nk", 3.044, 0.01640),
            ("gmx_benzene_VDW_u_nk", -3.033, 0.03438),
            ("gmx_expanded_ensemble_case_1", 75.993, 0.11056),
            ("gmx_expanded_ensemble_case_2", 76.009, 0.11220),
            ("gmx_expanded_ensemble_case_3", 76.219, 0.08886),
            ("gmx_water_particle_with_total_energy", -11.675, 0.065055),
            ("gmx_water_particle_with_potential_energy", -11.724, 0.064964),
            ("gmx_water_particle_without_energy", -11.660, 0.064914),
            ("amber_bace_example_complex_vdw", 2.39294, 0.051192),
            ("namd_tyr2ala", 11.0044, 0.10235),
            ("namd_idws", 0.221147, 0.041003),
            ("namd_idws_restarted", 7.081127, 0.0344211),
            ("namd_idws_restarted_reversed", -4.18405, 0.03457),
            ("gomc_benzene_u_nk", -0.87095, 0.071263),
        ],
    )
    def X_delta_f(self, request):
        get_unk, E, dE = request.param
        return alchemlyb.concat(request.getfixturevalue(get_unk)), E, dE

    def test_bar(self, X_delta_f):
        self.compare_delta_f(X_delta_f)

    def get_delta_f(self, est):
        ee = 0.0

        for i in range(len(est.d_delta_f_) - 1):
            ee += est.d_delta_f_.values[i][i + 1] ** 2
        # Use .iloc[0, -1] as we want to cater for both
        # delta_f_.loc[0.0, 1.0] and delta_f_.loc[(0.0, 0.0), (0.0, 1.0)]
        return est.delta_f_.iloc[0, -1], ee**0.5


class Test_Units:
    """Test the units."""

    @staticmethod
    @pytest.fixture()
    def u_nk(gmx_benzene_Coulomb_u_nk):
        return alchemlyb.concat(gmx_benzene_Coulomb_u_nk)

    def test_bar(self, u_nk):
        bar = BAR().fit(u_nk)
        assert bar.delta_f_.attrs["temperature"] == 300
        assert bar.delta_f_.attrs["energy_unit"] == "kT"
        assert bar.d_delta_f_.attrs["temperature"] == 300
        assert bar.d_delta_f_.attrs["energy_unit"] == "kT"

    def test_mbar(self, u_nk):
        mbar = MBAR().fit(u_nk)
        assert mbar.delta_f_.attrs["temperature"] == 300
        assert mbar.delta_f_.attrs["energy_unit"] == "kT"
        assert mbar.d_delta_f_.attrs["temperature"] == 300
        assert mbar.d_delta_f_.attrs["energy_unit"] == "kT"


class TestEstimatorMixOut:
    """Ensure that the attribute d_delta_f_, delta_f_, states_ cannot be
    modified."""

    @pytest.mark.parametrize("estimator", [MBAR, BAR])
    def test_d_delta_f_(self, estimator):
        _estimator = estimator()
        with pytest.raises(AttributeError):
            _estimator.d_delta_f_ = 1

    @pytest.mark.parametrize("estimator", [MBAR, BAR])
    def test_delta_f_(self, estimator):
        _estimator = estimator()
        with pytest.raises(AttributeError):
            _estimator.delta_f_ = 1

    @pytest.mark.parametrize("estimator", [MBAR, BAR])
    def test_states_(self, estimator):
        _estimator = estimator()
        with pytest.raises(AttributeError):
            _estimator.states_ = 1


def test_delta_f_columns(
    gmx_benzene_Coulomb_u_nk,
    gmx_expanded_ensemble_case_1,
):
    """Ensure columns are tuples when appropriate."""

    bar_1lambda = BAR().fit(alchemlyb.concat(gmx_benzene_Coulomb_u_nk))
    assert set(bar_1lambda.delta_f_.columns) == set([0.0, 0.25, 0.5, 0.75, 1.0])

    bar_4lambda = BAR().fit(alchemlyb.concat(gmx_expanded_ensemble_case_1))
    assert set(bar_4lambda.delta_f_.columns) == set(
        [
            (0.0, 0.1, 0.0, 0.0),
            (0.0, 0.4, 0.0, 0.0),
            (0.0, 1.0, 0.4, 0.002),
            (0.0, 1.0, 0.0, 0.0001),
            (0.0, 1.0, 0.1, 0.0002),
            (0.0, 0.84, 0.0, 0.0),
            (0.0, 0.68, 0.0, 0.0),
            (0.0, 1.0, 0.84, 0.2),
            (0.0, 1.0, 0.3, 0.001),
            (0.0, 1.0, 0.2, 0.0004),
            (0.0, 0.16, 0.0, 0.0),
            (0.0, 1.0, 0.52, 0.01),
            (0.0, 1.0, 0.92, 0.4),
            (0.0, 0.76, 0.0, 0.0),
            (0.0, 0.46, 0.0, 0.0),
            (0.0, 1.0, 0.6, 0.02),
            (0.0, 0.92, 0.0, 0.0),
            (0.0, 0.6, 0.0, 0.0),
            (0.0, 0.34, 0.0, 0.0),
            (0.0, 1.0, 0.76, 0.1),
            (0.0, 1.0, 1.0, 1.0),
            (0.0, 0.05, 0.0, 0.0),
            (0.0, 1.0, 0.48, 0.004),
            (0.0, 0.0, 0.0, 0.0),
            (0.0, 0.22, 0.0, 0.0),
            (0.0, 0.52, 0.0, 0.0),
            (0.0, 1.0, 0.68, 0.04),
            (0.0, 0.28, 0.0, 0.0),
        ]
    )


def test_bootstrap(gmx_benzene_Coulomb_u_nk):
    u_nk = alchemlyb.concat(gmx_benzene_Coulomb_u_nk)
    mbar = MBAR(n_bootstraps=2)
    mbar.fit(u_nk)
    mbar_bootstrap_mean = mbar.delta_f_.loc[0.00, 1.00]
    mbar_bootstrap_err = mbar.d_delta_f_.loc[0.00, 1.00]

    mbar = MBAR()
    mbar.fit(u_nk)
    mbar_mean = mbar.delta_f_.loc[0.00, 1.00]
    mbar_err = mbar.d_delta_f_.loc[0.00, 1.00]

    assert mbar_bootstrap_mean == mbar_mean
    assert mbar_bootstrap_err != mbar_err


def test_enthalpy_entropy_mbar(gmx_benzene_Coulomb_u_nk):
    """For the time being, only numpy <2.3 is supported due to potential changes in
    output of up to 0.1 kT (see PR Fix test #431 and issue numpy 2.3.0 madness
    choderalab/pymbar#556 )"""
    mbar = MBAR()
    u_nk = alchemlyb.concat(gmx_benzene_Coulomb_u_nk)
    mbar.fit(u_nk, compute_entropy_enthalpy=True)

    assert mbar.delta_f_.iloc[0, :].to_numpy() == pytest.approx(
        np.array([0.0, 1.619069, 2.557990, 2.986302, 3.041156]), abs=1e-6
    )
    assert mbar.delta_h_.iloc[0, :].to_numpy() == pytest.approx(
        np.array([0.0, 1.241970, 1.695000, 1.706555, 1.388906]), abs=1e-6
    )
    assert mbar.delta_sT_.iloc[0, :].to_numpy() == pytest.approx(
        np.array([0.0, -0.377099, -0.862990, -1.279746, -1.652250]), abs=1e-6
    )
    assert mbar.d_delta_f_.iloc[0, :].to_numpy() == pytest.approx(
        np.array([0.0, 0.008802, 0.014432, 0.018097, 0.020879]), abs=1e-6
    )
    assert mbar.d_delta_h_.iloc[0, :].to_numpy() == pytest.approx(
        np.array([0.0, 0.011598, 0.016538, 0.018077, 0.018940]), abs=1e-6
    )
    assert mbar.d_delta_sT_.iloc[0, :].to_numpy() == pytest.approx(
        np.array([0.0, 0.012242, 0.019519, 0.023606, 0.026107]), abs=1e-6
    )


def test_wrong_initial_f_k():
    with pytest.raises(
        ValueError, match="Only `BAR` is supported as string input to `initial_f_k`"
    ):
        MBAR(initial_f_k="aaa")


@pytest.mark.parametrize("initial_f_k", ["BAR", None])
def test_initial_f_k(gmx_benzene_Coulomb_u_nk, initial_f_k):
    u_nk = alchemlyb.concat(gmx_benzene_Coulomb_u_nk)
    mbar = MBAR(initial_f_k=initial_f_k)
    mbar.fit(u_nk)
    assert np.isclose(mbar.delta_f_.loc[0.00, 1.00], 3.0411556983908046)



import matplotlib.pyplot as plt
import numpy as np
from matplotlib.font_manager import FontProperties as FP
from matplotlib.axes import Axes
import pandas as pd

from ..postprocessors.units import get_unit_converter


def plot_convergence(
    dataframe: pd.DataFrame,
    units: None | str = None,
    final_error: None | float = None,
    ax: None | Axes = None,
) -> Axes:
    """Plot the forward and backward convergence.

     The input could be the result from
     :func:`~alchemlyb.convergence.forward_backward_convergence` or
     :func:`~alchemlyb.convergence.fwdrev_cumavg_Rc`. The input should be a
     :class:`pandas.DataFrame` which has column `Forward`, `Backward` and
     :attr:`pandas.DataFrame.attrs` should compile with :ref:`note-on-units`.
     The errorbar will be plotted if column `Forward_Error` and `Backward_Error`
     is present.

     `Forward`: A column of free energy estimate from the first X% of data,
     where optional `Forward_Error` column is the corresponding error.

     `Backward`: A column of free energy estimate from the last X% of data.,
     where optional `Backward_Error` column is the corresponding error.

    `final_error` is the error of the final value and is shown as the error band around the
    final value. It can be provided in case an estimate is available that is more appropriate
    than the default, which is the error of the last value in `Backward`.

     Parameters
     ----------
     dataframe : Dataframe
         Output Dataframe has column `Forward`, `Backward` or optionally
         `Forward_Error`, `Backward_Error` see :ref:`plot_convergence <plot_convergence>`.
     units : str
         The unit of the estimate. The default is `None`, which is to use the
         unit in the input. Setting this will change the output unit.
     final_error : float
         The error of the final value in ``units``. If not given, takes the last
         error in `backward_error`.
     ax : matplotlib.axes.Axes
         Matplotlib axes object where the plot will be drawn on. If ``ax=None``,
         a new axes will be generated.

     Returns
     -------
     matplotlib.axes.Axes
         An axes with the forward and backward convergence drawn.

     Note
     ----
     The code is taken and modified from
     `Alchemical Analysis <https://github.com/MobleyLab/alchemical-analysis>`_.


     .. versionchanged:: 1.0.0
         Keyword arg final_error for plotting a horizontal error bar.
         The array input has been deprecated.
         The units default to `None` which uses the units in the input.
     .. versionchanged:: 0.6.0
         data now takes in dataframe
     .. versionadded:: 0.4.0

    """
    if units is not None:
        dataframe = get_unit_converter(units)(dataframe)
    forward = dataframe["Forward"].to_numpy(dtype=float)
    if "Forward_Error" in dataframe:
        forward_error = dataframe["Forward_Error"].to_numpy(dtype=float)
    else:
        forward_error = np.zeros(len(forward))
    backward = dataframe["Backward"].to_numpy(dtype=float)
    if "Backward_Error" in dataframe:
        backward_error = dataframe["Backward_Error"].to_numpy(dtype=float)
    else:
        backward_error = np.zeros(len(backward))

    if ax is None:  # pragma: no cover
        fig, ax = plt.subplots(figsize=(8, 6))

    plt.setp(ax.spines["bottom"], color="#D2B9D3", lw=3, zorder=-2)
    plt.setp(ax.spines["left"], color="#D2B9D3", lw=3, zorder=-2)

    for dire in ["top", "right"]:
        ax.spines[dire].set_color("none")

    ax.xaxis.set_ticks_position("bottom")
    ax.yaxis.set_ticks_position("left")

    f_ts = np.linspace(0, 1, len(forward) + 1)[1:]
    r_ts = np.linspace(0, 1, len(backward) + 1)[1:]

    if final_error is None:
        final_error = backward_error[-1]

    if np.isfinite(backward[-1]) and np.isfinite(final_error):
        ax.fill_between(
            [0, 1],
            backward[-1] - final_error,
            backward[-1] + final_error,
            color="#D2B9D3",
            zorder=1,
        )
    line1 = ax.errorbar(
        f_ts,
        forward,
        yerr=forward_error,
        color="#736AFF",
        lw=3,
        zorder=2,
        marker="o",
        mfc="w",
        mew=2.5,
        mec="#736AFF",
        ms=12,
    )
    line2 = ax.errorbar(
        r_ts,
        backward,
        yerr=backward_error,
        color="#C11B17",
        lw=3,
        zorder=3,
        marker="o",
        mfc="w",
        mew=2.5,
        mec="#C11B17",
        ms=12,
    )

    xticks_spacing = len(r_ts) // 10 or 1
    xticks = r_ts[::xticks_spacing]
    plt.xticks(xticks, [f"{i:.2f}" for i in xticks], fontsize=10)
    plt.yticks(fontsize=10)

    ax.legend(
        (line1[0], line2[0]),
        ("Forward", "Reverse"),
        loc="best",
        prop=FP(size=18),
        frameon=False,
    )
    ax.set_xlabel(r"Fraction of the simulation time", fontsize=16, color="#151B54")
    ax.set_ylabel(r"$\Delta G$ ({})".format(units), fontsize=16, color="#151B54")
    plt.tick_params(axis="x", color="#D2B9D3")
    plt.tick_params(axis="y", color="#D2B9D3")
    plt.tight_layout()
    return ax


def plot_block_average(
    dataframe: pd.DataFrame,
    units: None | str = None,
    final_error: None | float = None,
    ax: None | Axes = None,
) -> Axes:
    """Plot the forward and backward convergence.

     The input could be the result from
     :func:`~alchemlyb.convergence.forward_backward_convergence` or
     :func:`~alchemlyb.convergence.fwdrev_cumavg_Rc`. The input should be a
     :class:`pandas.DataFrame` which has column `FE` and
     :attr:`pandas.DataFrame.attrs` should compile with :ref:`note-on-units`.
     The errorbar will be plotted if column `FE_Error` and `Backward_Error`
     is present.

     `FE`: A column of free energy estimate from some X% block of the data,
     where optional `FE_Error` column is the corresponding error.

    `final_error` is the error of the final value and is shown as the error band around the
    final value. It can be provided in case an estimate is available that is more appropriate
    than the default, which is the error of the last value in `Backward`.

     Parameters
     ----------
     dataframe : Dataframe
         Output Dataframe has column `Forward`, `Backward` or optionally
         `Forward_Error`, `Backward_Error` see :ref:`plot_convergence <plot_convergence>`.
     units : str
         The unit of the estimate. The default is `None`, which is to use the
         unit in the input. Setting this will change the output unit.
     final_error : float
         The error (standard deviation) of the final value in ``units``. If not given, takes the
         overall error of the time blocks, unless these were not provided, it which case it
         equals 1 kT.
     ax : matplotlib.axes.Axes
         Matplotlib axes object where the plot will be drawn on. If ``ax=None``,
         a new axes will be generated.

     Returns
     -------
     matplotlib.axes.Axes
         An axes with the forward and backward convergence drawn.

     Note
     ----
     The code is taken and modified from
     `Alchemical Analysis <https://github.com/MobleyLab/alchemical-analysis>`_.

     .. versionadded:: 2.4.0

    """
    if units is not None:
        dataframe = get_unit_converter(units)(dataframe)
    df_avg = dataframe["FE"].to_numpy(dtype=float)
    if "FE_Error" in dataframe:
        df_avg_error = dataframe["FE_Error"].to_numpy(dtype=float)
    else:
        df_avg_error = np.zeros(len(df_avg))

    if ax is None:  # pragma: no cover
        fig, ax = plt.subplots(figsize=(8, 6))

    plt.setp(ax.spines["bottom"], color="#D2B9D3", lw=3, zorder=-2)
    plt.setp(ax.spines["left"], color="#D2B9D3", lw=3, zorder=-2)

    for dire in ["top", "right"]:
        ax.spines[dire].set_color("none")

    ax.xaxis.set_ticks_position("bottom")
    ax.yaxis.set_ticks_position("left")

    f_ts = np.linspace(0, 1, len(df_avg) + 1)[1:]

    if final_error is None:
        if np.sum(df_avg_error) != 0:
            final_error = np.std(df_avg)
        else:
            final_error = 1.0

    if final_error is not None and np.isfinite(final_error):
        ax.fill_between(
            [0, 1],
            np.mean(df_avg) - final_error,
            np.mean(df_avg) + final_error,
            color="#D2B9D3",
            zorder=1,
        )
    ax.errorbar(
        f_ts,
        df_avg,
        yerr=df_avg_error,
        color="#736AFF",
        lw=3,
        zorder=2,
        marker="o",
        mfc="w",
        mew=2.5,
        mec="#736AFF",
        ms=12,
        label="Avg FE",
    )

    xticks_spacing = len(f_ts) // 10 or 1
    xticks = f_ts[::xticks_spacing]
    plt.xticks(xticks, [f"{i:.2f}" for i in xticks], fontsize=10)
    plt.yticks(fontsize=10)

    ax.legend(
        loc="best",
        prop=FP(size=18),
        frameon=False,
    )
    ax.set_xlabel(r"Fraction of the Simulation Time", fontsize=16, color="#151B54")
    ax.set_ylabel(r"$\Delta G$ ({})".format(units), fontsize=16, color="#151B54")
    plt.tick_params(axis="x", color="#D2B9D3")
    plt.tick_params(axis="y", color="#D2B9D3")
    plt.tight_layout()
    return ax



import alchemlyb


def test_version():
    try:
        version = alchemlyb.__version__
    except AttributeError:
        raise AssertionError("alchemlyb does not have __version__")

    assert len(version) > 0


def test_version__version__():
    import alchemlyb._version

    assert alchemlyb.__version__ == alchemlyb._version.__version__



import io

import pytest
from alchemtest.gmx import load_expanded_ensemble_case_1

from alchemlyb.parsing.util import anyopen


def test_gzip():
    """Test that gzip reads .gz files in the correct (text) mode."""
    dataset = load_expanded_ensemble_case_1()

    for leg in dataset["data"]:
        for filename in dataset["data"][leg]:
            with anyopen(filename, "r") as f:
                assert type(f.readline()) is str


def test_gzip_stream():
    """Test that `anyopen` reads streams with specified compression."""
    dataset = load_expanded_ensemble_case_1()

    for leg in dataset["data"]:
        for filename in dataset["data"][leg]:
            with open(filename, "rb") as f:
                with anyopen(f, mode="r", compression="gzip") as f_uc:
                    assert type(f_uc.readline()) is str


def test_gzip_stream_wrong():
    """Test that `anyopen` gives failure for attempting to decompress gzip
    stream with bz2.

    """
    dataset = load_expanded_ensemble_case_1()

    for leg in dataset["data"]:
        for filename in dataset["data"][leg]:
            with open(filename, "rb") as f:
                with anyopen(f, mode="r", compression="bzip2") as f_uc:
                    with pytest.raises(OSError, match="Invalid data stream"):
                        assert type(f_uc.readline()) is str


def test_gzip_stream_wrong_no_compression():
    """Test that `anyopen` gives passthrough when no compression specified on a
    stream.

    """
    dataset = load_expanded_ensemble_case_1()

    for leg in dataset["data"]:
        for filename in dataset["data"][leg]:
            with open(filename, "rb") as f:
                with anyopen(f, mode="r") as f_uc:
                    assert type(f_uc.readline()) is bytes


@pytest.mark.parametrize("extension", ["bz2", "gz"])
def test_file_roundtrip(extension, tmp_path):
    """Test that roundtripping write/read to a file works with `anyopen`."""

    data = "my momma told me to pick the very best one and you are not it"

    filepath = tmp_path / f"testfile.txt.{extension}"
    with anyopen(filepath, mode="w") as f:
        f.write(data)

    with anyopen(filepath, "r") as f:
        data_out = f.read()

    assert data_out == data


@pytest.mark.parametrize("extension,compression", [("bz2", "gzip"), ("gz", "bzip2")])
def test_file_roundtrip_force_compression(extension, compression, tmp_path):
    """Test that roundtripping write/read to a file works with `anyopen`,
    in which we force compression despite different extension.

    """

    data = "my momma told me to pick the very best one and you are not it"

    filepath = tmp_path / f"testfile.txt.{extension}"
    with anyopen(filepath, mode="w", compression=compression) as f:
        f.write(data)

    with anyopen(filepath, "r", compression=compression) as f:
        data_out = f.read()

    assert data_out == data


@pytest.mark.parametrize("compression", ["bzip2", "gzip"])
def test_stream_roundtrip(compression):
    """Test that roundtripping write/read to a stream works with `anyopen`"""

    data = "my momma told me to pick the very best one and you are not it"

    with io.BytesIO() as stream:
        # write to stream
        with anyopen(stream, mode="w", compression=compression) as f:
            f.write(data)

        # start at the beginning
        stream.seek(0)

        # read from stream
        with anyopen(stream, "r", compression=compression) as f:
            data_out = f.read()

        assert data_out == data


def test_stream_unsupported_compression():
    """Test that we throw a ValueError when an unsupported compression is used."""

    compression = "fakez"

    data = b"my momma told me to pick the very best one and you are not it"

    with io.BytesIO() as stream:
        # write to stream
        stream.write(data)

        # start at the beginning
        stream.seek(0)

        # read from stream
        with pytest.raises(ValueError):
            with anyopen(stream, "r", compression=compression) as f:
                f.read()



from .convergence import (
    forward_backward_convergence,
    fwdrev_cumavg_Rc,
    A_c,
    block_average,
)

__all__ = [
    "forward_backward_convergence",
    "fwdrev_cumavg_Rc",
    "A_c",
    "block_average",
]



import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pytest

import alchemlyb
from alchemlyb.convergence import forward_backward_convergence
from alchemlyb.estimators import MBAR, TI, BAR
from alchemlyb.visualisation import plot_convergence, plot_block_average
from alchemlyb.visualisation.dF_state import plot_dF_state
from alchemlyb.visualisation.mbar_matrix import plot_mbar_overlap_matrix
from alchemlyb.visualisation.ti_dhdl import plot_ti_dhdl


def test_plot_mbar_omatrix(gmx_benzene_Coulomb_u_nk):
    """Just test if the plot runs"""
    u_nk_coul = alchemlyb.concat(gmx_benzene_Coulomb_u_nk)
    mbar_coul = MBAR()
    mbar_coul.fit(u_nk_coul)

    assert isinstance(
        plot_mbar_overlap_matrix(mbar_coul.overlap_matrix), matplotlib.axes.Axes
    )
    assert isinstance(
        plot_mbar_overlap_matrix(
            mbar_coul.overlap_matrix,
            [
                1,
            ],
        ),
        matplotlib.axes.Axes,
    )

    # Bump up coverage
    overlap_maxtrix = mbar_coul.overlap_matrix
    overlap_maxtrix[0, 0] = 0.0025
    overlap_maxtrix[-1, -1] = 0.9975
    assert isinstance(plot_mbar_overlap_matrix(overlap_maxtrix), matplotlib.axes.Axes)


def test_plot_ti_dhdl(gmx_benzene_Coulomb_dHdl, gmx_benzene_VDW_dHdl):
    """Just test if the plot runs"""
    dHdl_coul = alchemlyb.concat(gmx_benzene_Coulomb_dHdl)
    ti_coul = TI()
    ti_coul.fit(dHdl_coul)

    ax = plot_ti_dhdl(ti_coul)
    assert isinstance(ax, matplotlib.axes.Axes)
    plt.close(ax.figure)

    fig, ax = plt.subplots(figsize=(8, 6))
    assert isinstance(plot_ti_dhdl(ti_coul, ax=ax), matplotlib.axes.Axes)
    assert isinstance(plot_ti_dhdl(ti_coul, labels=["Coul"]), matplotlib.axes.Axes)
    assert isinstance(
        plot_ti_dhdl(ti_coul, labels=["Coul"], colors=["r"]), matplotlib.axes.Axes
    )
    plt.close(fig)

    dHdl_vdw = alchemlyb.concat(gmx_benzene_VDW_dHdl)
    ti_vdw = TI().fit(dHdl_vdw)
    ax = plot_ti_dhdl([ti_coul, ti_vdw])
    assert isinstance(ax, matplotlib.axes.Axes)
    plt.close(ax.figure)

    ti_coul.dhdl = pd.DataFrame.from_dict(
        {"fep": range(100)}, orient="index", columns=np.arange(100) / 100
    ).T
    ti_coul.dhdl.attrs = dHdl_vdw.attrs
    ax = plot_ti_dhdl(ti_coul)
    assert isinstance(ax, matplotlib.axes.Axes)
    plt.close(ax.figure)


def test_plot_dF_state(
    gmx_benzene_Coulomb_dHdl,
    gmx_benzene_Coulomb_u_nk,
    gmx_benzene_VDW_u_nk,
    gmx_benzene_VDW_dHdl,
):
    """Just test if the plot runs"""
    u_nk_coul = alchemlyb.concat(gmx_benzene_Coulomb_u_nk)
    dHdl_coul = alchemlyb.concat(gmx_benzene_Coulomb_dHdl)
    u_nk_vdw = alchemlyb.concat(gmx_benzene_VDW_u_nk)
    dHdl_vdw = alchemlyb.concat(gmx_benzene_VDW_dHdl)

    ti_coul = TI().fit(dHdl_coul)
    ti_vdw = TI().fit(dHdl_vdw)
    bar_coul = BAR().fit(u_nk_coul)
    bar_vdw = BAR().fit(u_nk_vdw)
    mbar_coul = MBAR().fit(u_nk_coul)
    mbar_vdw = MBAR().fit(u_nk_vdw)

    dhdl_data = [
        (ti_coul, ti_vdw),
        (bar_coul, bar_vdw),
        (mbar_coul, mbar_vdw),
    ]
    fig = plot_dF_state(dhdl_data, orientation="portrait")
    assert isinstance(fig, matplotlib.figure.Figure)
    plt.close(fig)

    fig = plot_dF_state(dhdl_data, orientation="landscape")
    assert isinstance(fig, matplotlib.figure.Figure)
    plt.close(fig)

    fig = plot_dF_state(dhdl_data, labels=["MBAR", "TI", "BAR"])
    assert isinstance(fig, matplotlib.figure.Figure)
    plt.close(fig)

    with pytest.raises(ValueError):
        fig = plot_dF_state(
            dhdl_data,
            labels=[
                "MBAR",
                "TI",
            ],
        )

    fig = plot_dF_state(dhdl_data, colors=["#C45AEC", "#33CC33", "#F87431"])
    assert isinstance(fig, matplotlib.figure.Figure)
    plt.close(fig)

    with pytest.raises(ValueError):
        fig = plot_dF_state(dhdl_data, colors=["#C45AEC", "#33CC33"])

    with pytest.raises(ValueError):
        fig = plot_dF_state(dhdl_data, orientation="xxx")

    fig = plot_dF_state(ti_coul, orientation="landscape")
    assert isinstance(fig, matplotlib.figure.Figure)
    plt.close(fig)

    fig = plot_dF_state(ti_coul, orientation="portrait")
    assert isinstance(fig, matplotlib.figure.Figure)
    plt.close(fig)

    fig = plot_dF_state([ti_coul, bar_coul])
    assert isinstance(fig, matplotlib.figure.Figure)
    plt.close(fig)

    fig = plot_dF_state([(ti_coul, ti_vdw)])
    assert isinstance(fig, matplotlib.figure.Figure)
    plt.close(fig)


def test_plot_convergence_dataframe(gmx_benzene_Coulomb_u_nk):
    df = forward_backward_convergence([gmx_benzene_Coulomb_u_nk[0]], "MBAR")
    ax = plot_convergence(df)
    assert isinstance(ax, matplotlib.axes.Axes)
    plt.close(ax.figure)


def test_plot_convergence_dataframe_noerr():
    # Test the input from R_c
    data = pd.DataFrame(
        data={
            "Forward": range(100),
            "Backward": range(100),
            "data_fraction": np.linspace(0, 1, 100),
        }
    )
    data.attrs = {"temperature": 300, "energy_unit": "kT"}
    ax = plot_convergence(data, final_error=2)
    assert isinstance(ax, matplotlib.axes.Axes)
    plt.close(ax.figure)


def test_plot_convergence(gmx_benzene_Coulomb_u_nk):
    data_list = gmx_benzene_Coulomb_u_nk
    forward = []
    forward_error = []
    backward = []
    backward_error = []
    num_points = 10
    for i in range(1, num_points + 1):
        # Do the forward
        slice = int(len(data_list[0]) / num_points * i)
        u_nk_coul = alchemlyb.concat([data[:slice] for data in data_list])
        estimate = MBAR().fit(u_nk_coul)
        forward.append(estimate.delta_f_.loc[0.0, 1.0])
        forward_error.append(estimate.d_delta_f_.loc[0.0, 1.0])
        # Do the backward
        u_nk_coul = alchemlyb.concat([data[-slice:] for data in data_list])
        estimate = MBAR().fit(u_nk_coul)
        backward.append(estimate.delta_f_.loc[0.0, 1.0])
        backward_error.append(estimate.d_delta_f_.loc[0.0, 1.0])

    df = pd.DataFrame(
        data={
            "Forward": forward,
            "Forward_Error": forward_error,
            "Backward": backward,
            "Backward_Error": backward_error,
        }
    )
    df.attrs = estimate.delta_f_.attrs
    ax = plot_convergence(df)
    assert isinstance(ax, matplotlib.axes.Axes)
    plt.close(ax.figure)


def test_plot_convergence_final_nan():
    """Test the case where the Error of the final estimate is NaN."""
    df = pd.DataFrame(
        data={
            "Forward": [1, 2],
            "Forward_Error": [np.nan, np.nan],
            "Backward": [1, 2],
            "Backward_Error": [np.nan, np.nan],
        }
    )
    df.attrs = {"temperature": 300, "energy_unit": "kT"}
    ax = plot_convergence(df)
    assert isinstance(ax, matplotlib.axes.Axes)
    plt.close(ax.figure)


def test_plot_block_average(gmx_benzene_Coulomb_u_nk):
    data_list = gmx_benzene_Coulomb_u_nk
    fe = []
    fe_error = []
    num_points = 10
    for i in range(1, num_points + 1):
        slice = int(len(data_list[0]) / num_points * i)
        u_nk_coul = alchemlyb.concat([data[:slice] for data in data_list])
        estimate = MBAR().fit(u_nk_coul)
        fe.append(estimate.delta_f_.loc[0, 1])
        fe_error.append(estimate.d_delta_f_.loc[0, 1])

    df = pd.DataFrame(
        data={
            "FE": fe,
            "FE_Error": fe_error,
        }
    )
    df.attrs = estimate.delta_f_.attrs
    ax = plot_block_average(df)
    assert isinstance(ax, matplotlib.axes.Axes)
    plt.close(ax.figure)

    ax = plot_block_average(df, units="kJ/mol")
    assert isinstance(ax, matplotlib.axes.Axes)
    plt.close(ax.figure)

    df = df.drop("FE_Error", axis=1)
    ax = plot_block_average(df)
    assert isinstance(ax, matplotlib.axes.Axes)
    plt.close(ax.figure)

    ax = plot_block_average(df, final_error=1)
    assert isinstance(ax, matplotlib.axes.Axes)
    plt.close(ax.figure)

    ax = plot_block_average(df, final_error=np.inf)
    assert isinstance(ax, matplotlib.axes.Axes)
    plt.close(ax.figure)


class Test_Units:
    @staticmethod
    @pytest.fixture()
    def estimaters(gmx_benzene_Coulomb_dHdl, gmx_benzene_Coulomb_u_nk):
        dHdl_coul = alchemlyb.concat(gmx_benzene_Coulomb_dHdl)
        ti = TI().fit(dHdl_coul)
        u_nk_coul = alchemlyb.concat(gmx_benzene_Coulomb_u_nk)
        mbar = MBAR().fit(u_nk_coul)

        return ti, mbar

    @staticmethod
    @pytest.fixture(scope="class")
    def convergence():
        df = pd.DataFrame(
            data={
                "Forward": range(10),
                "Forward_Error": range(10),
                "Backward": range(10),
                "Backward_Error": range(10),
            }
        )
        df.attrs = {"temperature": 300, "energy_unit": "kT"}
        return df

    @pytest.mark.parametrize("units", [None, "kT", "kJ/mol", "kcal/mol"])
    def test_plot_dF_state(self, estimaters, units):
        fig = plot_dF_state(estimaters, units=units)
        assert isinstance(fig, matplotlib.figure.Figure)
        plt.close(fig)

    def test_plot_dF_state_unknown(self, estimaters):
        with pytest.raises(ValueError):
            plot_dF_state(estimaters, units="ddd")

    @pytest.mark.parametrize("units", [None, "kT", "kJ/mol", "kcal/mol"])
    def test_plot_ti_dhdl(self, estimaters, units):
        ti, mbar = estimaters
        ax = plot_ti_dhdl(ti, units=units)
        assert isinstance(ax, matplotlib.axes.Axes)
        plt.close(ax.figure)

    def test_plot_ti_dhdl_unknown(self, estimaters):
        ti, mbar = estimaters
        with pytest.raises(ValueError):
            plot_ti_dhdl(ti, units="ddd")

    @pytest.mark.parametrize("units", [None, "kT", "kJ/mol", "kcal/mol"])
    def test_plot_convergence(self, convergence, units):
        ax = plot_convergence(convergence)
        assert isinstance(ax, matplotlib.axes.Axes)
        plt.close(ax.figure)



import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator
import numpy.typing as npt
from .. import concat
from .base import _EstimatorMixOut


class TI_GQ(BaseEstimator, _EstimatorMixOut):
    """Thermodynamic integration (TI) with gaussian quadrature estimation.

    Parameters
    ----------

    verbose : bool, optional
        Set to True if verbose debug output is desired.

    Attributes
    ----------

    delta_f_ : DataFrame
        The estimated cumulative free energy from one state to another.

    d_delta_f_ : DataFrame
        The estimated statistical uncertainty (one standard deviation) in
        dimensionless cumulative free energies.

    states_ : list
        Lambda states for which free energy estimation were obtained.

    dhdl : DataFrame
        The estimated dhdl of each state.


    .. versionadded:: 2.1.0

    """

    special_points = {
        1: {"lambdas": [0.5], "weights": [1.0]},
        2: {"lambdas": [0.21132, 0.78867], "weights": [0.5, 0.5]},
        3: {"lambdas": [0.1127, 0.5, 0.88729], "weights": [0.27777, 0.44444, 0.27777]},
        4: {
            "lambdas": [0.06943, 0.33001, 0.66999, 0.93057],
            "weights": [0.17393, 0.32607, 0.32607, 0.17393],
        },
        5: {
            "lambdas": [0.04691, 0.23076, 0.5, 0.76923, 0.95308],
            "weights": [0.11846, 0.23931, 0.28444, 0.23931, 0.11846],
        },
        6: {
            "lambdas": [0.03377, 0.1694, 0.38069, 0.61931, 0.8306, 0.96623],
            "weights": [0.08566, 0.18038, 0.23396, 0.23396, 0.18038, 0.08566],
        },
        7: {
            "lambdas": [0.02544, 0.12923, 0.29707, 0.5, 0.70292, 0.87076, 0.97455],
            "weights": [0.06474, 0.13985, 0.19091, 0.20897, 0.19091, 0.13985, 0.06474],
        },
        8: {
            "lambdas": [
                0.01986,
                0.10167,
                0.23723,
                0.40828,
                0.59172,
                0.76277,
                0.89833,
                0.98014,
            ],
            "weights": [
                0.05061,
                0.11119,
                0.15685,
                0.18134,
                0.18134,
                0.15685,
                0.11119,
                0.05061,
            ],
        },
        9: {
            "lambdas": [
                0.01592,
                0.08198,
                0.19331,
                0.33787,
                0.5,
                0.66213,
                0.80669,
                0.91802,
                0.98408,
            ],
            "weights": [
                0.04064,
                0.09032,
                0.13031,
                0.15617,
                0.16512,
                0.15617,
                0.13031,
                0.09032,
                0.04064,
            ],
        },
        10: {
            "lambdas": [
                0.01305,
                0.06747,
                0.1603,
                0.2833,
                0.42556,
                0.57444,
                0.7167,
                0.8397,
                0.93253,
                0.98695,
            ],
            "weights": [
                0.03334,
                0.07473,
                0.10954,
                0.13463,
                0.14776,
                0.14776,
                0.13463,
                0.10954,
                0.07473,
                0.03334,
            ],
        },
        11: {
            "lambdas": [
                0.01089,
                0.05647,
                0.13492,
                0.24045,
                0.36523,
                0.5,
                0.63477,
                0.75955,
                0.86508,
                0.94353,
                0.98911,
            ],
            "weights": [
                0.02783,
                0.06279,
                0.09315,
                0.1166,
                0.1314,
                0.13646,
                0.1314,
                0.1166,
                0.09315,
                0.06279,
                0.02783,
            ],
        },
        12: {
            "lambdas": [
                0.00922,
                0.04794,
                0.11505,
                0.20634,
                0.31608,
                0.43738,
                0.56262,
                0.68392,
                0.79366,
                0.88495,
                0.95206,
                0.99078,
            ],
            "weights": [
                0.02359,
                0.05347,
                0.08004,
                0.10158,
                0.11675,
                0.12457,
                0.12457,
                0.11675,
                0.10158,
                0.08004,
                0.05347,
                0.02359,
            ],
        },
        13: {
            "lambdas": [
                0.00791,
                0.0412,
                0.09921,
                0.17883,
                0.27575,
                0.38477,
                0.5,
                0.61523,
                0.72425,
                0.82117,
                0.90079,
                0.9588,
                0.99209,
            ],
            "weights": [
                0.02024,
                0.04606,
                0.06944,
                0.08907,
                0.10391,
                0.11314,
                0.11628,
                0.11314,
                0.10391,
                0.08907,
                0.06944,
                0.04606,
                0.02024,
            ],
        },
        14: {
            "lambdas": [
                0.00686,
                0.03578,
                0.0864,
                0.15635,
                0.24238,
                0.34044,
                0.44597,
                0.55403,
                0.65956,
                0.75762,
                0.84365,
                0.9136,
                0.96422,
                0.99314,
            ],
            "weights": [
                0.01756,
                0.04008,
                0.06076,
                0.0786,
                0.09277,
                0.1026,
                0.10763,
                0.10763,
                0.1026,
                0.09277,
                0.0786,
                0.06076,
                0.04008,
                0.01756,
            ],
        },
        15: {
            "lambdas": [
                0.006,
                0.03136,
                0.0759,
                0.13779,
                0.21451,
                0.30292,
                0.3994,
                0.5,
                0.6006,
                0.69708,
                0.78549,
                0.86221,
                0.9241,
                0.96864,
                0.994,
            ],
            "weights": [
                0.01538,
                0.03518,
                0.05358,
                0.06979,
                0.08313,
                0.09308,
                0.09922,
                0.10129,
                0.09922,
                0.09308,
                0.08313,
                0.06979,
                0.05358,
                0.03518,
                0.01538,
            ],
        },
        16: {
            "lambdas": [
                0.0053,
                0.02771,
                0.06718,
                0.1223,
                0.19106,
                0.27099,
                0.3592,
                0.45249,
                0.54751,
                0.6408,
                0.72901,
                0.80894,
                0.8777,
                0.93282,
                0.97229,
                0.9947,
            ],
            "weights": [
                0.01358,
                0.03113,
                0.04758,
                0.06231,
                0.0748,
                0.08458,
                0.0913,
                0.09473,
                0.09473,
                0.0913,
                0.08458,
                0.0748,
                0.06231,
                0.04758,
                0.03113,
                0.01358,
            ],
        },
    }

    def __init__(self, verbose: bool = False) -> None:
        self.verbose = verbose

    def fit(self, dHdl: pd.DataFrame) -> "TI_GQ":
        """
        Compute cumulative free energy from one state to another by integrating
        dHdl across lambda values.

        Parameters
        ----------
        dHdl : DataFrame
            dHdl[n,k] is the potential energy gradient with respect to lambda
            for each configuration n and lambda k.

        """

        # sort by state so that rows from same state are in contiguous blocks,
        # and adjacent states are next to each other
        dHdl = dHdl.sort_index(level=dHdl.index.names[1:])  # type: ignore[arg-type]

        # obtain the mean and variance of the mean for each state
        # variance calculation assumes no correlation between points
        # used to calculate mean
        means = dHdl.groupby(level=dHdl.index.names[1:]).mean()
        variances = np.square(dHdl.groupby(level=dHdl.index.names[1:]).sem())

        weights = []
        # check if the lambdas in the simulations match the suggested values
        lambda_list, means_list, variances_list, index_list = (
            self.separate_mean_variance(means, variances)  # type: ignore[arg-type]
        )
        for lambdas in lambda_list:
            num_lambdas = len(lambdas)
            if num_lambdas not in self.special_points:
                raise ValueError(
                    f"TI_GQ only supports a set number of lambda windows ({list(self.special_points.keys())}) currently, \
                                 but {num_lambdas} lambda windows are given."
                )
            suggested_lambdas = self.special_points[num_lambdas]["lambdas"]
            if not np.allclose(lambdas, suggested_lambdas, rtol=0.1):
                raise ValueError(
                    f"lambda values, {suggested_lambdas}, are expected, but {lambdas} are given. Please use trapezoidal rule instead."
                )
            weights.extend(self.special_points[num_lambdas]["weights"])
        # means_new and variances_new are similar to means and variances, but with only values relevant to each lambda type (for multi-lambda situation)
        means_new = concat(means_list)
        mean_values = means_new.to_numpy()
        variances_new = concat(variances_list)
        variance_values = variances_new.to_numpy()

        # apply gaussian quadrature multiplication at each lambda state
        deltas = weights * mean_values
        deltas = np.insert(deltas, 0, [0.0], axis=0)  # type: ignore[assignment]
        deltas = np.append(deltas, [0.0], axis=0)  # type: ignore[assignment]
        d_deltas_squared = np.square(weights) * variance_values
        d_deltas_squared = np.insert(d_deltas_squared, 0, [0.0], axis=0)
        d_deltas_squared = np.append(d_deltas_squared, [0.0], axis=0)
        # build matrix of deltas between each state
        adelta = np.zeros((len(deltas), len(deltas)), dtype=float)
        ad_delta = np.zeros_like(adelta)

        for j in range(len(deltas)):
            out = []
            dout = []
            for i in range(len(deltas) - j):
                # Append cumulative free energy value from state i to i+j
                out.append(deltas[i] + deltas[i + 1 : i + j + 1].sum())  # type: ignore[attr-defined]
                # Append cumulative squared deviation of free energy from state i to i+j
                dout.append(
                    d_deltas_squared[i] + d_deltas_squared[i + 1 : i + j + 1].sum()
                )

            adelta += np.diagflat(np.array(out), k=j)
            ad_delta += np.diagflat(np.array(dout), k=j)

        adelta = adelta - adelta.T
        ad_delta = (ad_delta + ad_delta.T) - 2 * np.diagflat(d_deltas_squared)
        # yield standard delta_f_ cumulative free energies from one state to another
        self._delta_f_ = pd.DataFrame(adelta, columns=index_list, index=index_list)

        # yield standard deviation d_delta_f_ between each state
        self._d_delta_f_ = pd.DataFrame(
            np.sqrt(ad_delta),
            columns=index_list,
            index=index_list,
        )

        self.dhdl = means
        self.dhdl.attrs = dHdl.attrs
        self._states_ = means_new.index.values.tolist()
        self._delta_f_.attrs = dHdl.attrs
        self._d_delta_f_.attrs = dHdl.attrs

        return self

    @staticmethod
    def separate_mean_variance(
        means: pd.DataFrame, variances: pd.DataFrame
    ) -> tuple[
        list[pd.Index],
        list[pd.Series],
        list[pd.Series],
        list[float | tuple[float, ...]],
    ]:
        """
        For transitions with multiple lambda, the attr:`dhdl` would return
        a :class:`~pandas.DataFrame` which gives the dHdl for all the lambda
        states, regardless of whether it is perturbed or not. This function
        creates 3 lists of :class:`numpy.array`, :class:`pandas.Series` and
        :class:`pandas.Series` for each lambda, where the lists describe
        the lambda values, potential energy gradient and variance values for
        the lambdas state that is perturbed.

        Parameters
        ----------
        means: DataFrame
            means is the average potential energy gradient at each lambda.
        variances: DataFrame
            variances is variance of the potential energy gradient at each lambda.

        Returns
        ----------
        lambda_list : list
            A list of :class:`numpy.array` such that ``lambda_list[k]`` is the
            lambda values with respect to each type of lambda.
        dhdl_list : list
            A list of :class:`pandas.Series` such that ``dHdl_list[k]`` is the
            potential energy gradient with respect to lambda for each
            configuration that lambda k is perturbed.
        variance_list : list
            A list of :class:`pandas.Series` such that ``variance_list[k]`` is the
            variance of the potential energy gradient with respect to lambda for each
            configuration that lambda k is perturbed.
        index_list : list
            A list of :class:`float` or :class:`tuple` such that each :class:`float`
            or :class:`tuple` is the index of the final `delta_f_` and `d_delta_f_`
        """
        lambda_list: list[npt.NDArray] = []
        dhdl_list: list[pd.Series] = []
        variance_list: list[pd.Series] = []
        index_list: list[float | tuple[float]] = []
        # get the lambda names
        l_types = means.index.names
        # get the lambda vaules
        lambdas = means.reset_index()[means.index.names].values

        for i in range(len(l_types)):
            # obtain the lambda points between 0.0 and 1.0
            l_masks = (0.0 < lambdas[:, i]) & (lambdas[:, i] < 1.0)
            if not l_masks.any():
                continue
            new_means = means.iloc[l_masks, i]
            new_variances = variances.iloc[l_masks, i]
            index_list.extend(new_means.index)
            # for multi-lambda case, extract the relevant column
            for l_type in l_types:
                if l_type != l_types[i]:
                    new_means = new_means.reset_index(l_type, drop=True)
                    new_variances = new_variances.reset_index(l_type, drop=True)
            new_means.attrs = means.attrs
            new_variances.attrs = variances.attrs
            lambda_list.append(new_means.index)  # type: ignore[arg-type]
            dhdl_list.append(new_means)
            variance_list.append(new_variances)

        # add two end states at all lambda zeros and ones
        if len(l_types) == 1:
            index_list = [0.0] + index_list + [1.0]
        else:
            index_list = (
                [tuple([0.0] * len(l_types))]  # type: ignore[assignment]
                + index_list
                + [tuple([1.0] * len(l_types))]
            )

        return lambda_list, dhdl_list, variance_list, index_list  # type: ignore[return-value]



from functools import wraps
from typing import Any, Callable, Sequence

import pandas as pd

from ._version import __version__

__all__ = [
    "__version__",
]


def pass_attrs(
    func: Callable[..., pd.DataFrame | pd.Series],
) -> Callable[..., pd.DataFrame | pd.Series]:
    """Pass the attrs from the first positional argument to the output
    dataframe.


    .. versionadded:: 0.5.0
    """

    @wraps(func)
    def wrapper(
        input_dataframe: pd.DataFrame | pd.Series, *args: Any, **kwargs: Any
    ) -> pd.DataFrame | pd.Series:
        dataframe = func(input_dataframe, *args, **kwargs)
        dataframe.attrs = input_dataframe.attrs
        return dataframe

    return wrapper


def concat(
    objs: pd.DataFrame | pd.Series | Sequence[pd.DataFrame | pd.Series],
    *args: Any,
    **kwargs: Any,
) -> pd.DataFrame | pd.Series:
    """Concatenate pandas objects while persevering the attrs.

    Concatenate pandas objects along a particular axis with optional set
    logic along the other axes. If all pandas objects have the same  attrs
    attribute, the new pandas objects would have this attrs attribute. A
    ValueError would be raised if any pandas object has a different attrs.

    Parameters
    ----------
    objs
        A sequence or mapping of Series or DataFrame objects.

    Returns
    -------
    DataFrame
        Concatenated pandas object.

    Raises
    ------
    ValueError
        If not all pandas objects have the same attrs.

    See Also
    --------
    pandas.concat


    .. versionadded:: 0.5.0
    .. versionchanged:: 1.0.1
        When input is single dataframe, it will be sent out directly.

    """
    if isinstance(objs, (pd.DataFrame, pd.Series)):
        return objs
    # Sanity check
    try:
        attrs = objs[0].attrs
    except IndexError:  # except empty list as input
        raise ValueError("No objects to concatenate")

    for obj in objs:
        if attrs != obj.attrs:
            raise ValueError("All pandas objects should have the same attrs.")
    return pd.concat(objs, *args, **kwargs)



"""
Unit conversion and constants
"""

from scipy.constants import R, calorie
import pandas as pd
from typing import Callable

#: conversion factor from kJ to kcal, based on :data:`scipy.constants.calorie`
#: in :mod:`scipy.constants`
kJ2kcal = 1 / calorie

#: gas constant :math:`R` in kJ/(mol K), based on :data:`scipy.constants.R`
#: in :mod:`scipy.constants`
R_kJmol = R / 1000


def to_kT(
    df: pd.DataFrame | pd.Series, T: None | float = None
) -> pd.DataFrame | pd.Series:
    """Convert the unit of a DataFrame to `kT`.

    Note that if entropy values are passed it is assumed that they are
    multiplied by the temperature, S * T.

    If temperature `T` is not provided, the DataFrame need to have attribute
    `temperature` and `energy_unit`. Otherwise, the temperature of the output
    dateframe will be set accordingly.

    Parameters
    ----------
    df : DataFrame
        DataFrame to convert unit.
    T : float
        Temperature (default: None) in Kelvin.

    Returns
    -------
    DataFrame
        `df` converted.
    """
    new_df = df.copy()
    if T is not None:
        new_df.attrs["temperature"] = T
    elif "temperature" not in df.attrs:
        raise TypeError("Attribute temperature not found in the input Dataframe.")

    if "energy_unit" not in df.attrs:
        raise TypeError("Attribute energy_unit not found in the input Dataframe.")

    if df.attrs["energy_unit"] == "kT":
        return new_df
    elif df.attrs["energy_unit"] == "kJ/mol":
        new_df /= R_kJmol * df.attrs["temperature"]
        new_df.attrs["energy_unit"] = "kT"
        return new_df  # type: ignore[no-any-return]
    elif df.attrs["energy_unit"] == "kcal/mol":
        new_df /= R_kJmol * df.attrs["temperature"] * kJ2kcal
        new_df.attrs["energy_unit"] = "kT"
        return new_df  # type: ignore[no-any-return]
    else:
        raise ValueError(
            "energy_unit {} can only be kT, kJ/mol or kcal/mol.".format(
                df.attrs["energy_unit"]
            )
        )


def to_kcalmol(
    df: pd.DataFrame | pd.Series, T: None | float = None
) -> pd.DataFrame | pd.Series:
    """Convert the unit of a DataFrame to kcal/mol.

    Note that if entropy values are passed, the result is S * T in units
    of kcal/mol.

    If temperature `T` is not provided, the DataFrame need to have attribute
    `temperature` and `energy_unit`. Otherwise, the temperature of the output
    dateframe will be set accordingly.

    Parameters
    ----------
    df : DataFrame
        DataFrame to convert unit.
    T : float
        Temperature (default: None).

    Returns
    -------
    DataFrame
        `df` converted.
    """
    kt_df = to_kT(df, T)
    kt_df *= R_kJmol * df.attrs["temperature"] * kJ2kcal
    kt_df.attrs["energy_unit"] = "kcal/mol"
    return kt_df  # type: ignore[no-any-return]


def to_kJmol(
    df: pd.DataFrame | pd.Series, T: None | float = None
) -> pd.DataFrame | pd.Series:
    """Convert the unit of a DataFrame to kJ/mol.

    Note that if entropy values are passed, the result is S * T in units
    of kJ/mol.

    If temperature `T` is not provided, the DataFrame need to have attribute
    `temperature` and `energy_unit`. Otherwise, the temperature of the output
    dateframe will be set accordingly.

    Parameters
    ----------
    df : DataFrame
        DataFrame to convert unit.
    T : float
        Temperature (default: None).

    Returns
    -------
    DataFrame
        `df` converted.
    """
    kt_df = to_kT(df, T)
    kt_df *= R_kJmol * df.attrs["temperature"]
    kt_df.attrs["energy_unit"] = "kJ/mol"
    return kt_df  # type: ignore[no-any-return]


def get_unit_converter(units: str) -> Callable:
    """Obtain the converter according to the unit string.

    If `units` is 'kT', the `to_kT` converter is returned. If `units` is
    'kJ/mol', the `to_kJmol` converter is returned. If `units` is 'kcal/mol',
    the `to_kcalmol` converter is returned.

    Parameters
    ----------
    units : str
        The unit that the function converts to.

    Returns
    -------
    func
        converter


    .. versionadded:: 0.5.0
    """
    converters = {"kT": to_kT, "kJ/mol": to_kJmol, "kcal/mol": to_kcalmol}
    try:
        convert = converters[units]
    except KeyError:
        raise ValueError(
            f"Energy unit {units} is not supported, "
            f"choose one of {list(converters.keys())}"
        )
    return convert
